"""
Represents an VPC Security Group
"""
from footmark.vpc.vpcobject import TaggedVPCObject


class VSwitch(TaggedVPCObject):
    def __init__(self, connection=None):
        super(VSwitch, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'VSwitch:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.vswitch_id
        if name == 'subnet_id':
            return self.vswitch_id
        if name == 'name':
            return self.vswitch_name
        if name.startswith('subnet_'):
            return getattr(self, 'vswitch' + name[6:])
        raise AttributeError

    def __setattr__(self, name, value):
        if name == 'id':
            self.vswitch_id = value
        if name == 'name':
            self.vswitch_name = value
        if name == 'tags' and value:
            v = {}
            for tag in value['tag']:
                v[tag.get('TagKey')] = tag.get('TagValue', None)
            value = v
        if name.startswith('subnet_'):
            setattr(self, 'vswitch' + name[6:], value)
        super(TaggedVPCObject, self).__setattr__(name, value)

    def modify(self, name=None, description=None):
        params = {}
        if self.vswitch_name != name:
            params['vswitch_name'] = name
        if self.description != description:
            params['Descripton'] = description
        if params:
            params['vswitch_id'] = self.vswitch_id
            return self.connection.modify_vswitch(params)
        return False

    def get(self):
        return self.connection.get_vswitch_attribute(self.vswitch_id)

    def delete(self):
        return self.connection.delete_vswitch(self.vswitch_id)

    def read(self):
        vswitch = {}
        for name, value in self.__dict__.items():
            if name in ["connection", "region_id", "region"]:
                continue

            if name == 'vswitch_id':
                vswitch['id'] = value
                vswitch['subnet_id'] = value

            if name == 'status':
                name = 'state'
                value = str(value).lower()

            if name == 'zone_id':
                name = 'availability_zone'

            vswitch[name] = value
        return vswitch



# encoding: utf-8
"""
Represents a connection to the ECS service.
"""

import warnings
import time
import json
import base64
import logging
from footmark.ess.config import *
from footmark.connection import ACSQueryConnection
from footmark.exception import ECSResponseError
from footmark.resultset import ResultSet
from footmark.ess.configuration import ScalingConfiguration
from footmark.ess.group import ScalingGroup
from footmark.ess.instance import ScalingInstance
from footmark.ess.rule import ScalingRule
from footmark.ess.scheduled_task import ScheduledTask
# from aliyunsdkess.request.v20140828.DescribeScalingRulesRequest import


class ESSConnection(ACSQueryConnection):
    SDKVersion = '2014-08-28'
    DefaultRegionId = 'cn-hangzhou'
    DefaultRegionName = u'杭州'.encode("UTF-8")
    ResponseError = ECSResponseError

    def __init__(self, acs_access_key_id=None, acs_secret_access_key=None,
                 region=None, sdk_version=None, security_token=None, user_agent=None):
        """
        Init method to create a new connection to ECS.
        """
        if not region:
            region = self.DefaultRegionId
        self.region = region
        if sdk_version:
            self.SDKVersion = sdk_version

        self.ESSSDK = 'aliyunsdkess.request.v' + self.SDKVersion.replace('-', '')

        super(ESSConnection, self).__init__(acs_access_key_id=acs_access_key_id,
                                            acs_secret_access_key=acs_secret_access_key,
                                            region=self.region, product=self.ESSSDK,
                                            security_token=security_token, user_agent=user_agent)

    def build_filter_params(self, params, filters):
        if not isinstance(filters, dict):
            return

        flag = 1
        for key, value in filters.items():
            acs_key = key
            if acs_key.startswith('tag:'):
                while ('set_Tag%dKey' % flag) in params:
                    flag += 1
                if flag < 6:
                    params['set_Tag%dKey' % flag] = acs_key[4:]
                    params['set_Tag%dValue' % flag] = filters[acs_key]
                flag += 1
                continue
            if key == 'group_id':
                if not value.startswith('sg-') or len(value) != 12:
                    warnings.warn("The group-id filter now requires a security group "
                                  "identifier (sg-*) instead of a security group ID. "
                                  "The group-id " + value + "may be invalid.",
                                  UserWarning)
                params['set_SecurityGroupId'] = value
                continue
            if not isinstance(value, dict):
                acs_key = ''.join(s.capitalize() for s in acs_key.split('_'))
                params['set_' + acs_key] = value
                continue

            self.build_filters_params(params, value)

    def build_tags_params(self, params, tags, max_tag_number=None):
        tag_no = 1
        if tags:
            for key, value in tags.items():
                if tag_no < max_tag_number and key:
                    self.build_list_params(params, key, 'Tag' + bytes(tag_no+1) + 'Key')
                    self.build_list_params(params, value, 'Tag' + bytes(tag_no+1) + 'Value')
                    tag_no += 1

    def create_configuration(self, scaling_group_id, image_id, instance_type, security_group_id, name=None,
                             internet_charge_type=None, max_bandwidth_in=None, max_bandwidth_out=None,
                             system_disk_category='cloud_efficiency', system_disk_size=40, data_disks=None,
                             tags=None, key_pair_name=None, ram_role_name=None, user_data=None):
        """
        create an scaling configuration in ess

        :type str
        :param scaling_group_id: ID of the scaling group of a scaling configuration.
        :type str
        :param image_id: ID of an ECS instance image, indicating an image selected
        :type str
        :param instance_type: Resource type of an ECS instance.
        :type str
        :param security_group_id: ID of the security group to which a newly created instance belongs.
        :type str
        :param name: Name shown for the scheduled task. 
            The name must contain 2-40 English or Chinese characters, and start with a number, 
            a letter in upper or lower case or a Chinese character. 
            The name can contain numbers, “_“, “-“ or “.”. 
            The account name in the same scaling group is unique in the same region. 
            If this parameter value is not specified, the default value is ScalingConfigurationId.
        :type str
        :param internet_charge_type: Network billing type, Values: PayByBandwidth or PayByTraffic. Default to PayByBandwidth.
        :type str
        :param max_bandwidth_out: Maximum outgoing bandwidth from the public network, measured in Mbps (Mega bit per second). Default to 0.
        :type str
        :param max_bandwidth_in: Maximum incoming bandwidth from the public network, measured in Mbps (Mega bit per second). Default to 200.
        :type str
        :param system_disk_category: Category of the system disk. Values: cloud_efficiency or cloud_ssd. Default to cloud_efficiency.
        :type int
        :param system_disk_size: Size of the system disk. Values: 40~500. Default to 40.
        :type list
        :param data_disks: A list of hash/dictionaries of data disks. A maximum of four values can be entered.
            '[{size:"value", category:"value", snapshot_id:"value", delete_with_instance:"value"}]',
            size: Size of data disk. Values: 20 ~ 32768.
            category: Category of the data disk. Values: cloud_efficiency or cloud_ssd. Default to cloud_efficiency.
            snapshot_id: Snapshot used for creating the data disk. If it is specified, the size parameter is neglected, and the size of the created disk is the size of the snapshot.
            delete_with_instance: Whether the data disk will be released along with the instance. Default to true.
        :type list
        :param tags: A list of hash/dictionaries of instance
            tags, '[{tag_key:"value", tag_value:"value"}]', tag_key
            must be not null when tag_value isn't null
        :type str
        :param key_pair_name: Key pair name.
        :type str
        :param ram_role_name: The name of the instance RAM role.
        :type str
        :param user_data: The user-defined data of the instance. 
        
        :rtype: object
        :return: Returns a <footmark.ess.Configuration> object.

        """

        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')
        self.build_list_params(params, image_id, 'ImageId')
        self.build_list_params(params, instance_type, 'InstanceType')
        self.build_list_params(params, security_group_id, 'SecurityGroupId')

        if name:
            self.build_list_params(params, name, 'ScalingConfigurationName')
        if internet_charge_type:
            self.build_list_params(params, internet_charge_type, 'InternetChargeType')
        if max_bandwidth_in:
            self.build_list_params(params, max_bandwidth_in, 'InternetMaxBandwidthIn')
        if max_bandwidth_out:
            self.build_list_params(params, max_bandwidth_out, 'InternetMaxBandwidthOut')

        if system_disk_category:
            self.build_list_params(params, system_disk_category, 'SystemDisk.Category')
        if system_disk_size:
            self.build_list_params(params, system_disk_size, 'SystemDisk.Size')

        # Disks Details
        if data_disks:
            for i in range(len(data_disks)):
                if i > 3:
                    break
                disk = data_disks[i]
                if disk:
                    if 'size' in disk:
                        self.build_list_params(params, disk['size'], 'DataDisk' + bytes(i+1) + 'Size')
                    if 'category' in disk:
                        self.build_list_params(params, disk['category'], 'DataDisk' + bytes(i+1) + 'Category')
                    if 'snapshot_id' in disk:
                        self.build_list_params(params, disk['snapshot_id'], 'DataDisk' + bytes(i+1) + 'SnapshotId')
                    if 'delete_with_instance' in disk:
                        self.build_list_params(params, disk['delete_with_instance'],
                                               'DataDisk' + bytes(i+1) + 'DeleteWithInstance')

        self.build_tags_params(params, tags, max_tag_number=5)

        if key_pair_name:
            self.build_list_params(params, key_pair_name, 'KeyPairName')

        if user_data:
            self.build_list_params(params, base64.b64encode(user_data), 'UserData')

        if ram_role_name:
            self.build_list_params(params, ram_role_name, 'RamRoleName')
        instances = []

        result = self.get_object('CreateScalingConfiguration', params, ResultSet)
        return self.describe_configurations(scaling_group_id=scaling_group_id, scaling_configuration_ids=[result.scaling_configuration_id])[0]

    def describe_configurations(self, scaling_group_id=None, scaling_configuration_ids=None, scaling_configuration_names=None,
                                pagenumber=None, pagesize=50):
        """
        Retrieve all the scaling configurations 
        :type str
        :parameter scaling_group_id: ID of the scaling group of a scaling configuration.
        :type list
        :parameter scaling_configuration_ids: List ID of a scaling configurations.
        :type list
        :parameter scaling_configuration_names: List name of a scaling configurations.
        :type int
        :parameter pagenumber: Page number of the scaling configuration list. The initial value and default value are both 1.
        :type int
        :parameter pagesize: When querying by page, this parameter indicates the number of lines per page. Maximum value: 50; default value: 10.
        
        :rtype: list
        :return: A list of  :class:`footmark.ess.configuration`

        """

        cfgs = []
        params = {}
        if scaling_group_id:
            self.build_list_params(params, scaling_group_id, 'ScalingGroupId')
        if scaling_configuration_ids and len(scaling_configuration_ids) > 0:
            for i in range(len(scaling_configuration_ids)):
                if i < 10 and scaling_configuration_ids[i]:
                    self.build_list_params(params, scaling_configuration_ids[i], 'ScalingConfigurationId'+ bytes(i+1))
        if scaling_configuration_names:
            for i in range(len(scaling_configuration_names)):
                if i < 10 and scaling_configuration_names[i]:
                    self.build_list_params(params, scaling_configuration_names[i], 'ScalingConfigurationName'+ bytes(i+1))

        self.build_list_params(params, pagesize, 'PageSize')

        pNum = pagenumber
        if not pNum:
            pNum = 1
        while True:
            self.build_list_params(params, pNum, 'PageNumber')
            cfg_list = self.get_list('DescribeScalingConfigurations', params, ['ScalingConfigurations', ScalingConfiguration])

            for cfg in cfg_list:
                cfgs.append(cfg)
            if pagenumber or len(cfg_list) < pagesize:
                break
            pNum += 1

        return cfgs

    def terminate_configuration(self, scaling_configuration_id):
        """
        Deletes a specified scaling configuration.
        
        An active scaling configuration in a scaling group cannot be deleted.
        If any ECS instance created according to a scaling configuration is still in the scaling group, the scaling configuration cannot be deleted.

        :type str
        :param scaling_configuration_id: ID of a scaling configuration.

        :rtype: bool
        :return: The result of deleting scaling configuration.
        """
        params = {}

        self.build_list_params(params, scaling_configuration_id, 'ScalingConfigurationId')

        return self.get_status('DeleteScalingConfiguration', params)

    def create_group(self, max_size, min_size, name=None, default_cooldown=None, removal_policies=None,
                     load_balancer_ids=None, db_instance_ids=None, vswitch_ids=None):
        """
        A scaling group is a collection of ECS instances with the same application scenarios.
        It defines the maximum and minimum numbers of ECS instances in the group, 
        and their associated Server Load Balancer instances, RDS instances, and other attributes.

        :type int
        :param max_size: Maximum number of ECS instances in the scaling group. Value range: [0, 100].
        :type int
        :param min_size: Minimum number of ECS instances in the scaling group. Value range: [0, 100].
        :type str
        :param name: Name shown for the scaling group, which must contain 2-40 characters (English or Chinese).
            The name must begin with a number, an upper/lower-case letter or a Chinese character and may contain numbers, “_“, “-“ or “.”. 
            The account name is unique in the same region.
            If this parameter is not specified, the default value is its ID.
        :type int
        :param default_cooldown: Default cool-down time (in seconds) of the scaling group. Value range: [0, 86400]. Default to is 300.
        :type list
        :param removal_policies: Policy for removing ECS instances from the scaling group.
            Optional values:
                OldestInstance: removes the first ECS instance attached to the scaling group.
                NewestInstance: removes the first ECS instance attached to the scaling group.
                OldestScalingConfiguration: removes the ECS instance with the oldest scaling configuration.
            Default values: OldestScalingConfiguration and OldestInstance.
            At most 2 policies can be entered.
        :type list
        :param load_balancer_ids: ID list of a Server Load Balancer instance.
        :type list
        :param db_instance_ids: ID list of an RDS instance.
        :type list
        :param vswitch_ids: ID list of a VSwitch. It is used to create instance in multiple zones.
            At most 5 VSwitches in a VPC can be specified.
            The priority of VSwitches descends from 1 to 5, and 1 indicates the highest priority.
            When you fail to create an instance in the zone to which a specified VSwitch belongs, 
            another VSwitch with less priority replaces the specified one automatically. 
        
        :rtype: object
        :return: Returns a <footmark.ess.Group> object.

        """

        params = {}

        self.build_list_params(params, max_size, 'MaxSize')
        if min_size > max_size:
            self.build_list_params(params, max_size, 'MinSize')
        else:
            self.build_list_params(params, min_size, 'MinSize')

        if name:
            self.build_list_params(params, name, 'ScalingGroupName')
        if default_cooldown:
            self.build_list_params(params, default_cooldown, 'DefaultCooldown')
        if removal_policies:
            for i in range(len(removal_policies)):
                if i < 2 and removal_policies[i]:
                    self.build_list_params(params, removal_policies[i], 'RemovalPolicy' + bytes(i + 1));
        if load_balancer_ids:
            self.build_list_params(params, json.dumps(load_balancer_ids), 'LoadBalancerIds')

        if db_instance_ids:
            self.build_list_params(params, json.dumps(db_instance_ids), 'DBInstanceIds')

        if vswitch_ids:
            self.build_list_params(params, vswitch_ids, 'VSwitchIds')

        result = self.get_object('CreateScalingGroup', params, ResultSet)
        return self.describe_groups(scaling_group_ids=[result.scaling_group_id])[0]

    def describe_groups(self, scaling_group_ids=None, scaling_group_names=None, pagenumber=None, pagesize=50):
        """
        Query the information of a scaling group. Scaling groups have the following life cycle states:

            Active: In this state, the scaling group can receive scaling rule execution requests and trigger scaling activities.
            Inactive: In this state, the scaling group does not receive scaling rule execution requests.
            Deleting: The scaling group is being deleted and does not receive scaling rule execution requests. 
        
        :type list
        :param scaling_group_ids: List ID of a scaling groups.
            At most 20 IDs can be entered.
            Invalid scaling group IDs are not displayed in query results, and no error is reported.
        :type list
        :param scaling_group_names: List name of a scaling groups.
            At most 20 IDs can be entered.
            Invalid scaling group IDs are not displayed in query results, and no error is reported.
        :type int
        :param pagenumber: Page number of the scaling group list. The initial value and default value are both 1.
        :type int
        :param pagesize: When querying by page, this parameter indicates the number of lines per page. Maximum value: 50; default value: 10.
        
        :rtype: list
        :return: A list of :class:`footmark.ess.group`

        """

        cfgs = []
        params = {}
        if scaling_group_ids:
            for i in range(len(scaling_group_ids)):
                if i < 20 and scaling_group_ids[i]:
                    self.build_list_params(params, scaling_group_ids[i], 'ScalingGroupId' + bytes(i + 1))
        if scaling_group_names:
            for i in range(len(scaling_group_names)):
                if i < 20 and scaling_group_names[i]:
                    self.build_list_params(params, scaling_group_names[i], 'ScalingGroupName' + bytes(i + 1))

        self.build_list_params(params, pagesize, 'PageSize')

        pNum = pagenumber
        if not pNum:
            pNum = 1
        while True:
            self.build_list_params(params, pNum, 'PageNumber')
            cfg_list = self.get_list('DescribeScalingGroups', params, ['ScalingGroups', ScalingGroup])
            for cfg in cfg_list:
                cfgs.append(cfg)
            if pagenumber or len(cfg_list) < pagesize:
                break
            pNum += 1

        return cfgs

    def enable_group(self, scaling_group_id, scaling_configuration_id=None, instance_ids=None):
        """
        Enables the specified scaling group.
        After the scaling group is successfully enabled (the group is active), the ECS instances specified by the interface are attached to the group.

        ForceDelete indicates whether to forcibly delete a scaling group and remove and release ECS instances if the scaling group has ECS instances or scaling activities are in progress.

        Restrictions on attaching ECS instances:
        
            1. The attached ECS instance and the scaling group must be in the same region.
            2. The attached ECS instance and the instance with active scaling configurations must be of the same type.
            3. The attached ECS instance must in the running state.
            4. The attached ECS instance has not been attached to other scaling groups.
            5. The attached ECS instance supports Subscription and Pay-As-You-Go payment methods.
            6. If the VswitchID is specified for a scaling group, you cannot attach Classic ECS instances or ECS instances on other VPCs to the scaling group.
            7. If the VswitchID is not specified for the scaling group, ECS instances of the VPC type cannot be attached to the scaling group
        
        :type str
        :param scaling_group_id: ID of a scaling group.
        :type str
        :param scaling_configuration_id: ID of a active scaling configuration.
        :type list
        :param instance_ids: ID of the ECS instance to be attached to the scaling group after it is enabled.
            At most 20 IDs can be entered.

        :rtype: bool
        :return: The result of enabling scaling group.
        """
        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')

        if scaling_configuration_id:
            self.build_list_params(params, scaling_configuration_id, 'ActiveScalingConfigurationId')

        if instance_ids:
            for i in range(len(instance_ids)):
                if i < 20 and instance_ids[i]:
                    self.build_list_params(params, instance_ids[i], 'InstanceId' + bytes(i + 1))

        return self.get_status('EnableScalingGroup', params)

    def disable_group(self, scaling_group_id):
        """
        Disable a specified scaling group.

        The scaling activities in progress before the scaling group is disabled are continued until completion, 
        whereas scaling activities triggered after the scaling group is disabled are rejected.
        
        :type str
        :param scaling_group_id: ID of a scaling group.
        :type str
        
        :rtype: bool
        :return: The result of disabling scaling group.
        """
        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')

        return self.get_status('DisableScalingGroup', params)

    def modify_group(self, scaling_group_id, max_size=None, min_size=None, name=None, default_cooldown=None,
                     removal_policies=None, scaling_configuration_id=None):

        """
        Modifies the attributes of a scaling group.
        
        The interface can be called only when the scaling group is active or inactive.

        When the scaling configuration specified for the scaling group needs to be modified,
        the instance type attribute of the modified scaling configuration must be consistent with that of the active scaling configuration.

        After a new scaling configuration is added to the scaling group, the running ECS instances which are created based on the previous scaling configuration remain unchanged.
        When the number (total capacity) of ECS instances in the scaling group does not meet the modified MaxSize or MinSize specification,
        the Auto Scaling service automatically attaches or removes ECS instances to/from the group to make odds even.
    
        
        :type str
        :param scaling_group_id: ID of a scaling group.
        :type int
        :param max_size: Maximum number of ECS instances in the scaling group. Value range: [0, 100].
        :type int
        :param min_size: Minimum number of ECS instances in the scaling group. Value range: [0, 100].
        :type str
        :param name: Name shown for the scaling group, which must contain 2-40 characters (English or Chinese).
            The name must begin with a number, an upper/lower-case letter or a Chinese character and may contain numbers, “_“, “-“ or “.”. 
            The account name is unique in the same region.
            If this parameter is not specified, the default value is its ID.
        :type int
        :param default_cooldown: Default cool-down time (in seconds) of the scaling group. Value range: [0, 86400]. Default to is 300.
        :type list
        :param removal_policies: Policy for removing ECS instances from the scaling group.
            Optional values:
                OldestInstance: removes the first ECS instance attached to the scaling group.
                NewestInstance: removes the first ECS instance attached to the scaling group.
                OldestScalingConfiguration: removes the ECS instance with the oldest scaling configuration.
            Default values: OldestScalingConfiguration and OldestInstance.
            At most 2 policies can be entered.
        :type str
        :param scaling_configuration_id: ID of a active scaling configuration.
        
        :rtype: object
        :return: Returns a <footmark.ess.Group> object.
    
        """

        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')

        if max_size is not None:
            self.build_list_params(params, max_size, 'MaxSize')
        if min_size is not None:
            self.build_list_params(params, min_size, 'MinSize')
        if name:
            self.build_list_params(params, name, 'ScalingGroupName')
        if default_cooldown:
            self.build_list_params(params, default_cooldown, 'DefaultCooldown')
        if removal_policies:
            for i in range(len(removal_policies)):
                if i < 2 and removal_policies[i]:
                    self.build_list_params(params, removal_policies[i], 'RemovalPolicy' + bytes(i + 1));
        if scaling_configuration_id:
            self.build_list_params(params, scaling_configuration_id, 'ActiveScalingConfigurationId')

        return self.get_status('ModifyScalingGroup', params)

    def terminate_group(self, scaling_group_id, force=False):
        """
        Delete a specified scaling group.

        ForceDelete indicates whether to forcibly delete a scaling group and remove and release ECS instances if the scaling group has ECS instances or scaling activities are in progress.

        When ForceDelete is set to false, the scaling group can be deleted only when the following conditions are met:

            Condition 1: No scaling activities are in progress in the scaling group.
            Condition 2: The current number (total capacity) of ECS instances in the scaling group is 0.
        When the two conditions are met, the scaling group is disabled and then deleted.

        When ForceDelete is set to true, the scaling group is disabled to reject new scaling activity requests. 
        When the existing scaling activity is completed, all ECS instances are removed from the scaling group and the group is then deleted (manually attached ECS instances are removed from the scaling group, whereas ECS instances automatically created by the Auto Scaling service are deleted).
        Deleting a scaling group also deletes scaling configurations, rules, activities, and requests.
        
        :type str
        :param scaling_group_id: ID of a scaling group.
        :type bool
        :param force: Indicates whether to forcibly delete a scaling group and remove and release ECS instances if the scaling group has ECS instances or scaling activities are in progress.
            Default to False.

        :rtype: bool
        :return: The result of deleting scaling group.
        """
        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')

        if force:
            self.build_list_params(params, force, 'ForceDelete')

        return self.get_status('DeleteScalingGroup', params)

    def attach_instances(self, scaling_group_id, instance_ids):
        """
        Attaches an ECS instance to a specified scaling group. Restrictions on the attached ECS instance:

        The attached ECS instance and the scaling group must be in the same region.
        The attached ECS instance must in the running state.
        The attached ECS instance has not been attached to other scaling groups.
        The attached ECS instance supports Subscription and Pay-As-You-Go payment methods.
        If the VswitchID is specified for a scaling group, you cannot attach Classic ECS instances or ECS instances on other VPCs to the scaling group.
        If the VswitchID is not specified for the scaling group, ECS instances of the VPC type cannot be attached to the scaling group.
        
        :type str
        :param scaling_group_id: ID of a scaling group.
        :type list
        :param instance_ids: ID of the ECS instance to be attached to the scaling group after it is enabled.
            At most 20 IDs can be entered.

        :rtype: bool
        :return: The result of deleting scaling group.
        """
        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')

        for i in range(len(instance_ids)):
                if i < 20 and instance_ids[i]:
                    self.build_list_params(params, instance_ids[i], 'InstanceId' + bytes(i + 1))

        result = self.get_object('AttachInstances', params, ResultSet)
        return self.wait_for_instances_status(scaling_group_id, instance_ids, InService, activity_id=result.scaling_activity_id,
                                              delay=DefaultWaitForInterval, timeout=DefaultTimeOut)

    def remove_instances(self, scaling_group_id, instance_ids):
        """
        Removes an ECS instance from a specified scaling group.

        When the ECS instance automatically created by the Auto Scaling service is removed from the scaling group, the ECS instance is disabled and released.
        When the manually attached ECS instance is removed from the scaling group, the ECS instance is neither disabled nor released.
        The interface can be called only when the scaling group is active.
        The interface can be called only when no scaling activity in the scaling group is in progress.
        When no scaling activity in the scaling group is in progress, the interface can be directly executed without cooldown.
        Successfully calling this interface only means that the Auto Scaling service has accepted the call request, and the scaling activity can be executed, but does not necessarily mean that the scaling activity can be successfully executed. You can use the returned ScalingActivityId to check the status of the scaling activity.
        When the total capacity of instances of the scaling group minus instances specified by this interface is smaller than than MinSize, the call fails.

        :type str
        :param scaling_group_id: ID of a scaling group.
        :type list
        :param instance_ids: ID of the ECS instance to be attached to the scaling group after it is enabled.
            At most 20 IDs can be entered.

        :rtype: bool
        :return: The result of deleting scaling group.
        """
        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')

        for i in range(len(instance_ids)):
            if i < 20 and instance_ids[i]:
                self.build_list_params(params, instance_ids[i], 'InstanceId' + bytes(i + 1))

        result = self.get_object('RemoveInstances', params, ResultSet)

        timeout = DefaultTimeOut
        while True:
            instances = self.describe_instances(scaling_group_id=scaling_group_id, instance_ids=instance_ids)
            if not instances:
                return True

            timeout -= 5

            if timeout <= 0:
                raise Exception("Timeout Error: Waiting for removing instances, time-consuming {0} seconds."
                                " Scaling activity ID: {1}.".format(DefaultTimeOut, result.scaling_activity_id))

            time.sleep(5)

        return False

    def describe_instances(self, scaling_group_id=None, scaling_configuration_id=None, instance_ids=None,
                           health_status=None, lifecycle_state=None, creation_type=None, pagenumber=None, pagesize=50):
        """
        Queries the list of ECS instances in a scaling group. 
        You can query by scaling group ID, scaling configuration ID, health status, lifecycle status, and creation type.
        
        :type str
        :param scaling_group_id: ID of a scaling group.
        :type str
        :param scaling_configuration_id: ID of a scaling configuration
        :type list
        :param instance_ids: ID of the ECS instance to be attached to the scaling group after it is enabled.
            At most 20 IDs can be entered.
        :type str
        :param health_status: Health status of an ECS instance in the scaling group. Options: Healthy and Unhealthy.
        :type str
        :param lifecycle_state: Lifecycle status of an ECS instance in the scaling group. Options: 
            - InService: the ECS instance has been added to the scaling group and runs properly. 
            - Pending: the ECS instance is being attached to the scaling group with relevant configurations not completed.
            - Removing: the ECS instance is being removed from the scaling group.
        :type str
        :param creation_type: ECS instance creation type. Options: 
            - AutoCreated: the ECS instance is automatically created by the Auto Scaling service in the scaling group.
            - Attached: the ECS instance is created outside the Auto Scaling service and manually attached to the scaling group.
        :type int
        :param pagenumber: Page number of the scaling group list. The initial value and default value are both 1.
        :type int
        :param pagesize: When querying by page, this parameter indicates the number of lines per page. Maximum value: 50; default value: 10.
        
        :rtype: list
        :return: A list of :class:`footmark.ess.instance`

        """

        instances = []
        params = {}
        if scaling_group_id:
            self.build_list_params(params, scaling_group_id, 'ScalingGroupId')
        if scaling_configuration_id:
            self.build_list_params(params, scaling_configuration_id, 'ScalingConfigurationId')
        if instance_ids:
            for i in range(len(instance_ids)):
                if i < 20 and instance_ids[i]:
                    self.build_list_params(params, instance_ids[i], 'InstanceId' + bytes(i + 1))
        if health_status:
            self.build_list_params(params, health_status, 'HealthStatus')
        if lifecycle_state:
            self.build_list_params(params, lifecycle_state, 'LifecycleState')
        if creation_type:
            self.build_list_params(params, creation_type, 'CreationType')

        self.build_list_params(params, pagesize, 'PageSize')

        pNum = pagenumber
        if not pNum:
            pNum = 1
        while True:
            self.build_list_params(params, pNum, 'PageNumber')
            cfg_list = self.get_list('DescribeScalingInstances', params, ['ScalingInstances', ScalingInstance])
            for cfg in cfg_list:
                instances.append(cfg)
            if pagenumber or len(cfg_list) < pagesize:
                break
            pNum += 1

        return instances

    def create_rule(self, scaling_group_id, adjustment_type, adjustment_value, name=None, cooldown=None):
        """
        A scaling rule defines specific scaling actions, for example, adding or removing n ECS instances.

        If the execution of a scaling rule results in a number of ECS instances in the scaling group that is less than the MinSize or greater than the MaxSize,
        Auto Scaling automatically adjusts the number of ECS instances to be added or removed by executing the “Adjust scaling group instance quantity to MinSize” 
        or “Adjust scaling group instance quantity to MaxSize” rule.

        Example 1: If a scaling group has a MaxSize of 3, the current number of instances (Total Capacity) is 2, 
                   and a scaling rule instructs the system to “add 3 ECS instances”, this operation actually only adds one ECS instance.
                  (The values in the scaling rule are not changed.)
        Example 2: If a scaling group has a MinSize of 2, the current number of instances (Total Capacity) is 3,
                   and a scaling rule instructs the system to “remove 5 ECS instances”, this operation actually only removes one ECS instance.
                   (The values in the scaling rule are not changed.)

        :type str
        :param scaling_group_id: ID of the scaling group of a scaling rule.
        :type str
        :param adjustment_type: Adjustment mode of a scaling rule. Optional values:
            - QuantityChangeInCapacity: It is used to increase or decrease a specified number of ECS instances.
            - PercentChangeInCapacity: It is used to increase or decrease a specified proportion of ECS instances.
            - TotalCapacity: It is used to adjust the quantity of ECS instances in the current scaling group to a specified value.
        :type int
        :param adjustment_value: Adjusted value of a scaling rule. Value range:
            - QuantityChangeInCapacity: (0, 100] U (-100, 0]
            - PercentChangeInCapacity: [0, 10000] U [-10000, 0]
            - TotalCapacity: [0, 100]
        :type str
        :param name: Name shown for the scaling rule, which is a string containing 2 to 40 English or Chinese characters.
                It must begin with a number, a letter (case-insensitive) or a Chinese character and can contain numbers, “_”, “-“ or “.”.
                The account name in the same scaling group is unique in the same region.
                If this parameter value is not specified, the default value is ScalingRuleId.
        :type int
        :param cooldown: Cool-down time of a scaling rule. Value range: [0, 86,400], in seconds. The default value is empty.
        
        :rtype: object
        :return: Returns a <footmark.ess.Rule> object.

        """

        params = {}

        self.build_list_params(params, scaling_group_id, 'ScalingGroupId')
        self.build_list_params(params, adjustment_type, 'AdjustmentType')
        self.build_list_params(params, adjustment_value, 'AdjustmentValue')

        if name:
            self.build_list_params(params, name, 'ScalingRuleName')
        if cooldown is not None:
            self.build_list_params(params, cooldown, 'Cooldown')

        result = self.get_object('CreateScalingRule', params, ResultSet)
        return self.describe_rules(scaling_group_id=scaling_group_id, scaling_rule_ids=[result.scaling_rule_id])[0]

    def describe_rules(self, scaling_group_id=None, scaling_rule_ids=None, scaling_rule_names=None, scaling_rule_aris=None,
                       pagenumber=None, pagesize=50):
        """
        Queries information of a scaling rule. You can query all scaling rules in a scaling group by specifying the scaling group ID.
        
        :type str
        :param scaling_group_id: ID of the scaling group of a scaling configuration.
        :type list
        :param scaling_rule_ids: List ID of a scaling configurations. A maximum of 10 values can be entered.
        :type list
        :param scaling_rule_names: List name of a scaling configurations. A maximum of 10 values can be entered.
        :type list
        :param scaling_rule_aris: Unique identifier of a scaling rule. A maximum of 10 values can be entered.
        :type int
        :param pagenumber: Page number of the scaling rule list. The initial value and default value are both 1.
        :type int
        :param pagesize: When querying by page, this parameter indicates the number of lines per page. Maximum value: 50;
            default value: 50.
        
        :rtype: list
        :return: A list of  :class:`footmark.ess.rule`
    
        """

        rules = []
        params = {}
        if scaling_group_id:
            self.build_list_params(params, scaling_group_id, 'ScalingGroupId')
        if scaling_rule_ids:
            for i in range(len(scaling_rule_ids)):
                if i < 10 and scaling_rule_ids[i]:
                    self.build_list_params(params, scaling_rule_ids[i], 'ScalingRuleId'+ bytes(i+1))
        if scaling_rule_names:
            for i in range(len(scaling_rule_names)):
                if i < 10 and scaling_rule_names[i]:
                    self.build_list_params(params, scaling_rule_names[i], 'ScalingRuleName'+bytes(i+1))
        if scaling_rule_aris:
            for i in range(len(scaling_rule_aris)):
                if i < 10 and scaling_rule_aris[i]:
                    self.build_list_params(params, scaling_rule_aris[i], 'ScalingRuleAri'+bytes(i+1))

        self.build_list_params(params, pagesize, 'PageSize')

        pNum = pagenumber
        if not pNum:
            pNum = 1
        while True:
            self.build_list_params(params, pNum, 'PageNumber')
            rule_list = self.get_list('DescribeScalingRules', params, ['ScalingRules', ScalingRule])
            # API DescribeScalingRules has not 'ScalingRules' when there is no any rule.
            if not isinstance(rule_list, list):
                return rules

            for rule in rule_list:
                rules.append(rule)
            if pagenumber or len(rule_list) < pagesize:
                break
            pNum += 1

        return rules

    def modify_rule(self, scaling_rule_id, adjustment_type=None, adjustment_value=None, name=None, cooldown=None):
        """
        Modifies the attributes of a scaling rule.

        :type str
        :param scaling_rule_id: ID of a scaling rule.
        :type str
        :param adjustment_type: Adjustment mode of a scaling rule. Optional values:
            - QuantityChangeInCapacity: It is used to increase or decrease a specified number of ECS instances.
            - PercentChangeInCapacity: It is used to increase or decrease a specified proportion of ECS instances.
            - TotalCapacity: It is used to adjust the quantity of ECS instances in the current scaling group to a specified value.
        :type int
        :param adjustment_value: Adjusted value of a scaling rule. Value range:
            - QuantityChangeInCapacity: (0, 100] U (-100, 0]
            - PercentChangeInCapacity: [0, 10000] U [-10000, 0]
            - TotalCapacity: [0, 100]
        :type str
        :param name: Name shown for the scaling rule, which is a string containing 2 to 40 English or Chinese characters.
                It must begin with a number, a letter (case-insensitive) or a Chinese character and can contain numbers, “_”, “-“ or “.”.
                The account name in the same scaling group is unique in the same region.
                If this parameter value is not specified, the default value is ScalingRuleId.
        :type int
        :param cooldown: Cool-down time of a scaling rule. Value range: [0, 86,400], in seconds. The default value is empty.
        
        :rtype: bool
        :return: The result of modifying.

        """

        params = {}

        self.build_list_params(params, scaling_rule_id, 'ScalingRuleId')

        if adjustment_type:
            self.build_list_params(params, adjustment_type, 'AdjustmentType')

        if adjustment_value is not None:
            self.build_list_params(params, adjustment_value, 'AdjustmentValue')

        if name:
            self.build_list_params(params, name, 'ScalingRuleName')
        if cooldown is not None:
            self.build_list_params(params, cooldown, 'Cooldown')

        return self.get_status('ModifyScalingRule', params)

    def terminate_rule(self, scaling_rule_id):
        """
        Deletes a specified scaling rule.

        :type str
        :param scaling_rule_id: ID of a scaling rule.
        
        :rtype: bool
        :return: The result of deleting.

        """

        params = {}

        self.build_list_params(params, scaling_rule_id, 'ScalingRuleId')

        return self.get_status('DeleteScalingRule', params)

    def create_scheduled_task(self, scaling_rule_ari, launch_time, name=None, description=None, launch_expiration_time=None,
                              recurrence_type=None, recurrence_value=None, recurrence_end_time=None, task_enabled=None):
        """
        This operation creates a scheduled task according to input parameters.

        You can create up to 20 scheduled tasks.
        
        When the trigger of a scheduled task fails because a scaling activity in a scaling group is in progress or the scaling group is disabled, 
        the scheduled task is automatically retried within the LaunchExpirationTime; otherwise, the scheduled trigger task is abandoned.
        
        If multiple tasks are scheduled at similar times to execute the rule of the same group, the earliest task triggers the scaling activity first, 
        and other tasks make attempts to execute the rule within their Launch Expiration Time because a scaling group executes only one scaling activity at a time. 
        If another scheduled task is still making triggering attempts within its Launch Expiration Time after the scaling activity is finished, 
        the scaling rule is executed and the corresponding scaling activity is triggered.

        :type str
        :param scaling_rule_ari: The unique identifier of the scaling rule.
        :type str
        :param launch_time: Time point at which the scheduled task is triggered.
            The date format follows the ISO8601 standard and uses UTC time. It is in the format of YYYY-MM-DDThh:mmZ.
            If RecurrenceType is specified, the time point specified by this attribute is the default time point at which the circle is executed. 
            If RecurrenceType is not specified, the task is executed once on the designated date and time.
            A time point 90 days after creation or modification cannot be entered.
        :type str
        :param name: Display name of the scheduled task, which must be 2-40 characters (English or Chinese) long. 
            It must begin with a number, an upper/lower-case letter or a Chinese character and may contain numbers, “_”, “-“ or “.”.
            The account name is unique in the same region.
            If this parameter is not specified, the default value ScheduledScalingTaskId is used.
        :type str
        :param description: Description of the scheduled task, which is 2-200 characters (English or Chinese) long.
        :type int
        :param launch_expiration_time: Time period within which the failed scheduled task is retried. The default value is 600s. Value range: [0, 21600]
        :type str
        :param recurrence_type: Type of the scheduled task to be repeated. Optional values:
            - Daily: Recurrence interval by day for a scheduled task.
            - Weekly: Recurrence interval by week for a scheduled task.
            - Monthly: Recurrence interval by month for a scheduled task.
            RecurrenceType, RecurrenceValue and RecurrenceEndTime must be specified at the same time.
        :type str
        :param recurrence_value: Value of the scheduled task to be repeated.
            - Daily: Only one value in the range [1,31] can be filled.
            - Weekly: Multiple values can be filled. The values of Sunday to Saturday are 0 to 6 in sequence. Multiple values shall be separated by a comma “,”.
            - Monthly: In the format of A-B. The value range of A and B is 1 to 31, and the B value must be greater than the A value.
            RecurrenceType, RecurrenceValue and RecurrenceEndTime must be specified at the same tiem.
        :type str
        :param recurrence_end_time: End time of the scheduled task to be repeated.
            The date format follows the ISO8601 standard and uses UTC time. It is in the format of YYYY-MM-DDThh:mmZ.
            A time point 90 days after creation or modification cannot be entered.
            RecurrenceType, RecurrenceValue and RecurrenceEndTime must be specified at the same time.
        :type bool
        :param task_enabled: Whether to enable the scheduled task.
            - When the parameter is set to true, the task is enabled.
            - When the parameter is set to false, the task is disabled.
            The default value is true.
        
        :rtype: object
        :return: Returns a <footmark.ess.ScheduledTask> object.

        """

        params = {}

        self.build_list_params(params, scaling_rule_ari, 'ScheduledAction')
        self.build_list_params(params, launch_time, 'LaunchTime')

        if name:
            self.build_list_params(params, name, 'ScheduledTaskName')
        if description:
            self.build_list_params(params, description, 'Description')
        if launch_expiration_time is not None:
            self.build_list_params(params, launch_expiration_time, 'LaunchExpirationTime')
        if recurrence_type:
            self.build_list_params(params, recurrence_type, 'RecurrenceType')
        if recurrence_value:
            self.build_list_params(params, recurrence_value, 'RecurrenceValue')
        if recurrence_end_time:
            self.build_list_params(params, recurrence_end_time, 'RecurrenceEndTime')
        if task_enabled is True or task_enabled is None:
            self.build_list_params(params, True, 'TaskEnabled')

        result = self.get_object('CreateScheduledTask', params, ResultSet)
        return self.describe_scheduled_tasks(scheduled_task_ids=[result.scheduled_task_id])[0]

    def describe_scheduled_tasks(self, scheduled_task_ids=None, scheduled_task_names=None, scaling_rule_aris=None,
                                 pagenumber=None, pagesize=50):
        """
        Queries information of a scheduled task.

        You can use the Scaling Rule Ari to query the relevant scheduled task.
        
        :type list
        :parameter scheduled_task_ids: ID of the scheduled task. You can enter at most 20 IDs.
        :type list
        :parameter scheduled_task_names: Display name of the schedules task. You can enter at most 20 display names.
        :type list
        :parameter scaling_rule_aris: Unique identifier of a scaling rule. A maximum of 20 values can be entered.
        :type int
        :parameter pagenumber: Page number of the scaling scheduled task list. The initial value and default value are both 1.
        :type int
        :parameter pagesize: When querying by page, this parameter indicates the number of lines per page. Maximum value: 50; default value: 50.
        
        :rtype: list
        :return: A list of  :class:`footmark.ess.ScheduledTask`
    
        """

        tasks = []
        params = {}

        if scheduled_task_ids:
            for i in range(len(scheduled_task_ids)):
                if i < 20 and scheduled_task_ids[i]:
                    self.build_list_params(params, scheduled_task_ids[i], 'ScheduledTaskId'+ bytes(i+1))
        if scheduled_task_names:
            for i in range(len(scheduled_task_names)):
                if i < 20 and scheduled_task_names[i]:
                    self.build_list_params(params, scheduled_task_names[i], 'ScheduledTaskName'+ bytes(i+1))
        if scaling_rule_aris:
            for i in range(len(scaling_rule_aris)):
                if i < 10 and scaling_rule_aris[i]:
                    self.build_list_params(params, scaling_rule_aris[i], 'ScheduledAction'+ bytes(i+1))

        self.build_list_params(params, pagesize, 'PageSize')

        pNum = pagenumber
        if not pNum:
            pNum = 1
        while True:
            self.build_list_params(params, pNum, 'PageNumber')
            task_list = self.get_list('DescribeScheduledTasks', params, ['ScheduledTasks', ScheduledTask])

            for rule in task_list:
                tasks.append(rule)
            if pagenumber or len(task_list) < pagesize:
                break
            pNum += 1

        return tasks

    def modify_scheduled_task(self, scheduled_task_id, scaling_rule_ari=None, launch_time=None, name=None, description=None,
                              launch_expiration_time=None, recurrence_type=None, recurrence_value=None, recurrence_end_time=None, task_enabled=None):
        """
        Modifies the attributes of a scheduled task.

        :type str
        :param scheduled_task_id: ID of the scheduled task.
        :type str
        :param scaling_rule_ari: The unique identifier of the scaling rule.
        :type str
        :param launch_time: Time point at which the scheduled task is triggered.
            The date format follows the ISO8601 standard and uses UTC time. It is in the format of YYYY-MM-DDThh:mmZ.
            If RecurrenceType is specified, the time point specified by this attribute is the default time point at which the circle is executed. 
            If RecurrenceType is not specified, the task is executed once on the designated date and time.
            A time point 90 days after creation or modification cannot be entered.
        :type str
        :param name: Display name of the scheduled task, which must be 2-40 characters (English or Chinese) long. 
            It must begin with a number, an upper/lower-case letter or a Chinese character and may contain numbers, “_”, “-“ or “.”.
            The account name is unique in the same region.
            If this parameter is not specified, the default value ScheduledScalingTaskId is used.
        :type str
        :param description: Description of the scheduled task, which is 2-200 characters (English or Chinese) long.
        :type int
        :param launch_expiration_time: Time period within which the failed scheduled task is retried. The default value is 600s. Value range: [0, 21600]
        :type str
        :param recurrence_type: Type of the scheduled task to be repeated. Optional values:
            - Daily: Recurrence interval by day for a scheduled task.
            - Weekly: Recurrence interval by week for a scheduled task.
            - Monthly: Recurrence interval by month for a scheduled task.
            RecurrenceType, RecurrenceValue and RecurrenceEndTime must be specified at the same time.
        :type str
        :param recurrence_value: Value of the scheduled task to be repeated.
            - Daily: Only one value in the range [1,31] can be filled.
            - Weekly: Multiple values can be filled. The values of Sunday to Saturday are 0 to 6 in sequence. Multiple values shall be separated by a comma “,”.
            - Monthly: In the format of A-B. The value range of A and B is 1 to 31, and the B value must be greater than the A value.
            RecurrenceType, RecurrenceValue and RecurrenceEndTime must be specified at the same tiem.
        :type str
        :param recurrence_end_time: End time of the scheduled task to be repeated.
            The date format follows the ISO8601 standard and uses UTC time. It is in the format of YYYY-MM-DDThh:mmZ.
            A time point 90 days after creation or modification cannot be entered.
            RecurrenceType, RecurrenceValue and RecurrenceEndTime must be specified at the same time.
        :type bool
        :param task_enabled: Whether to enable the scheduled task.
            - When the parameter is set to true, the task is enabled.
            - When the parameter is set to false, the task is disabled.
            The default value is true.
        
        :rtype: object
        :return: Returns a <footmark.ess.ScheduledTask> object.

        """

        params = {}

        self.build_list_params(params, scheduled_task_id, 'ScheduledTaskId')
        if scaling_rule_ari:
            self.build_list_params(params, scaling_rule_ari, 'ScalingRuleAri')
        if launch_time:
            self.build_list_params(params, launch_time, 'LaunchTime')
        if name:
            self.build_list_params(params, name, 'ScheduledTaskName')
        if description:
            self.build_list_params(params, description, 'Description')
        if launch_expiration_time is not None:
            self.build_list_params(params, launch_expiration_time, 'LaunchExpirationTime')
        if recurrence_type:
            self.build_list_params(params, recurrence_type, 'RecurrenceType')
        if recurrence_value:
            self.build_list_params(params, recurrence_value, 'RecurrenceValue')
        if recurrence_end_time:
            self.build_list_params(params, recurrence_end_time, 'RecurrenceEndTime')
        if task_enabled is True or task_enabled is None:
            self.build_list_params(params, True, 'TaskEnabled')

        return self.get_status('ModifyScheduledTask', params)

    def terminate_scheduled_task(self, scheduled_task_id):
        """
        Deletes a specified scheduled task.

        :type str
        :param scheduled_task_id: ID of the scheduled task.
        
        :rtype: bool
        :return: The result of deleting.

        """

        params = {}

        self.build_list_params(params, scheduled_task_id, 'ScheduledTaskId')

        return self.get_status('DeleteScheduledTask', params)

    def wait_for_instances_status(self, scaling_group_id, instance_ids, status, activity_id=None, delay=DefaultWaitForInterval, timeout=DefaultTimeOut):
        """
        To verify instances status has become expected after attaching or detaching instances
        """
        tm = timeout
        try:
            while True:
                instances = self.describe_instances(scaling_group_id=scaling_group_id, instance_ids=instance_ids)
                success = 0
                for inst in instances:
                    if str(inst.status).lower() == str.lower(status):
                        success += 1

                if success == len(instance_ids):
                    return True

                tm -= delay

                if tm <= 0:
                    raise Exception("Timeout Error: Waiting for scaling instances status is %s, time-consuming {0} seconds."
                                    " Scaling activity ID: {1}.".format(DefaultTimeOut, activity_id))

                time.sleep(delay)

            return False
        except Exception as e:
            raise e


"""
Represents an ECS Network Interface
"""
from footmark.ecs.ecsobject import TaggedECSObject


class NetworkInterfaceSet(TaggedECSObject):
    """
    Represents an network interface.
    """

    def __init__(self, connection=None):
        super(NetworkInterfaceSet, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'NetworkInterface:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.network_interface_id
        if name == 'name':
            return self.network_interface_name
        if name == 'state':
            return self.status
        raise AttributeError("Object {0} does not have attribute {1}".format(self.__repr__(), name))

    def __setattr__(self, name, value):
        if name == 'id':
            self.network_interface_id = value
        if name == 'name':
            self.network_interface_name = value
        if name == 'status':
            value = value.lower()
        if name == 'state':
            self.status = value
        super(TaggedECSObject, self).__setattr__(name, value)

    def _update(self, updated):
        self.__dict__.update(updated.__dict__)

    def update(self, validate=False):
        """
        Update the network interface's state information by making a call to fetch
        the current interface attributes from the service.
        """
        rs = self.connection.get_network_interface(self.id)
        if rs:
            self._update(rs)
        elif validate:
            raise ValueError('%s is not a valid Network Interface ID' % self.id)
        return self.state

    def get(self):
        """
        Attach the network interface with an instance.
        """
        return self.connection.get_network_interface(self.id)

    def attach(self, instance_id):
        """
        Attach the network interface with an instance.
        """
        if instance_id and not self.instance_id:
            return self.connection.attach_network_interface(self.id, instance_id)
        return False

    def detach(self, instance_id):
        """
        Detach the network interface from an instance.
        """
        if self.instance_id and self.instance_id == instance_id:
            return self.connection.detach_network_interface(self.id, instance_id)
        return False

    def modify(self, security_group_ids=None, name=None, description=None):
        """
        Modify the network interface.
        """
        params = {}
        if security_group_ids and sorted(security_group_ids) != sorted(self.security_group_ids["security_group_id"]):
            params['security_group_ids'] = security_group_ids
        if name != self.name:
            params['network_interface_name'] = name
        if description != self.description:
            params['description'] = description
        if params:
            params['network_interface_id'] = self.id
            return self.connection.modify_network_interface(params)
        return False

    def delete(self):
        """
        Terminate the network interface
        """
        return self.connection.delete_network_interface(self.id)

    def read(self):
        eni = {}
        for name, value in self.__dict__.items():
            if name in ["connection", "region_id", "region"]:
                continue

            if name == "network_interface_id":
                eni['id'] = value

            if name == 'private_ip_sets':
                temp = []
                for ip in value["private_ip_set"]:
                    temp.append({
                        'private_ip_address': ip["private_ip_address"],
                        'primary_address': ip["primary"]
                    })
                name = "private_ip_addresses"
                value = temp

            if name == "security_group_ids":
                name = "security_groups"
                value = value['security_group_id']

            if name == 'network_interface_name':
                name = 'name'

            if name == 'status':
                name = 'state'
                value = str(value).lower()

            if name == 'zone_id':
                name = 'availability_zone'

            eni[name] = value
        return eni



# coding:utf-8
import sys
import time

reload(sys)
sys.setdefaultencoding("utf-8")
"""
Handles basic connections to ACS
"""

import footmark
import importlib
from footmark.exception import FootmarkServerError
from footmark.provider import Provider
import json
import yaml
from footmark.resultset import ResultSet

from pprint import pprint

from aliyunsdkcore import client
from aliyunsdkcore.acs_exception.exceptions import ServerException
# from aliyunsdkecs.request.v20140526.DescribeNetworkInterfacesRequest import



class ACSAuthConnection(object):
    def __init__(self, acs_access_key_id=None, acs_secret_access_key=None,
                 region=None, provider='acs', security_token=None, user_agent=None):
        """
        :keyword str acs_access_key_id: Your ACS Access Key ID (provided by
            Alicloud). If none is specified, the value in your
            ``ACS_ACCESS_KEY_ID`` environmental variable is used.
        :keyword str acs_secret_access_key: Your ACS Secret Access Key
            (provided by Alicloud). If none is specified, the value in your
            ``ACS_SECRET_ACCESS_KEY`` environmental variable is used.
        :keyword str security_token: The security token associated with
            temporary credentials issued by STS.  Optional unless using
            temporary credentials.  If none is specified, the environment
            variable ``ACS_SECURITY_TOKEN`` is used if defined.

        :keyword str region: The region ID.

        """
        self.region = region
        self.user_agent = user_agent
        if isinstance(provider, Provider):
            # Allow overriding Provider
            self.provider = provider
        else:
            self._provider_type = provider
            self.provider = Provider(self._provider_type,
                                     acs_access_key_id,
                                     acs_secret_access_key,
                                     security_token)

    def acs_access_key_id(self):
        return self.provider.access_key

    acs_access_key_id = property(acs_access_key_id)
    access_key = acs_access_key_id

    def acs_secret_access_key(self):
        return self.provider.secret_key

    acs_secret_access_key = property(acs_secret_access_key)
    secret_key = acs_secret_access_key

    def region_id(self):
        return self.region


class ACSQueryConnection(ACSAuthConnection):
    ResponseError = FootmarkServerError

    def __init__(self, acs_access_key_id=None, acs_secret_access_key=None, region=None,
                 product=None, security_token=None, provider='acs',
                 user_agent='Alicloud-Footmark-v'+footmark.__version__):

        super(ACSQueryConnection, self).__init__(
            acs_access_key_id,
            acs_secret_access_key,
            region=region,
            security_token=security_token,
            provider=provider,
            user_agent=user_agent)

        self.product = product
        self.user_agent = user_agent

    def import_request(self, action):
        try:
            target = importlib.import_module(self.product + '.' + action + 'Request')
            return getattr(target, action + 'Request')()
        except Exception as e:
            raise Exception("Importing {0} request got an error: {1}.".format(action, e))

    def build_request_params(self, filters):
        params = {}

        if not filters:
            return params

        if not isinstance(filters, dict):
            raise Exception("Error: 'filters' must be a dict. Current value is: {0}.".format(filters))

        if not filters.get('Action', filters.get('action')):
            raise Exception("'Action' is required for this request.")

        # get all of the specified request's method names dict, like {'InstanceId':'set_InstanceId', 'Name':'set_Name'}
        methods = {}
        request = self.import_request(filters.get('Action', filters.get('action')))
        if not request:
            raise Exception("There is no available request about action {0}.".format(filters.get('Action', filters.get('action'))))
        for name in dir(request):
            if str(name).startswith('set_') and name[4] <= 'Z':
                methods[str(name[4:]).lower()] = name

        # build request params dict, like {'set_InstanceId':'i-12345', 'set_Name':'abcd'}
        if methods:
            for key, value in filters.items():
                if value is None:
                    continue
                name = str(key).lower().replace("-", "").replace("_", "")
                if name in methods:
                    params[methods[name]] = value
                if key in ['Action', 'action']:
                    params['Action'] = value
        return params

    def make_request(self, action, params=None):
        conn = client.AcsClient(self.acs_access_key_id, self.acs_secret_access_key, self.region, user_agent=self.user_agent)
        if not conn:
            footmark.log.error('%s %s' % ('Null AcsClient ', conn))
            raise self.FootmarkClientError('Null AcsClient ', conn)

        timeout = 200
        delay = 3
        while timeout > 0:
            try:
                request = self.import_request(action)
                request.set_accept_format('json')
                if params and isinstance(params, dict):
                    for k, v in params.items():
                        if hasattr(request, k):
                            getattr(request, k)(v)
                        else:
                            request.add_query_param(k[4:], v)
                return conn.do_action_with_exception(request)
            except Exception as e:
                if str(e.error_code) == "SDK.ServerUnreachable" \
                        or str(e.message).__contains__("SDK.ServerUnreachable") \
                        or str(e.message).__contains__("Unable to connect server: timed out"):
                    time.sleep(delay)
                    timeout -= delay
                    continue
                raise e

        return None

    def build_list_params(self, params, items, label):
        if isinstance(items, str):
            items = str(items).strip()
        params['set_%s' % label] = items

    def parse_response(self, markers, body, connection):
        results = []
        body = yaml.safe_load(body)
        if not markers:
            markers = ["", ResultSet]

        result_set = ResultSet
        # For get_object
        if not markers[0] and markers[1] is not ResultSet:
            result_set = markers[1](connection)

        for key, value in body.items():
            self.parse_value(value)
            setattr(result_set, self.convert_name(key), value)

        if markers[0] and markers[0] in body:
            for value in getattr(result_set, self.convert_name(markers[0])).itervalues():
                if isinstance(value, list):
                    for sub_value in value:
                        element = markers[1](connection)
                        for k, v in sub_value.items():
                            setattr(element, k, v)
                        results.append(element)
                elif isinstance(value, dict):
                    element = markers[1](connection)
                    for k, v in value.items():
                        setattr(element, k, v)
                    results.append(element)
                else:
                    element = markers[1](connection)
                    setattr(element, k, v)
                    results.append(element)
            return results
        return result_set

    def parse_value(self, value):
        if isinstance(value, list):
            for item in value:
                self.parse_value(item)
        if isinstance(value, dict):
            for k, v in value.items():
                if isinstance(v, dict) or isinstance(v, list):
                    self.parse_value(v)
                else:
                    value.pop(k)
                    value[self.convert_name(k)] = v
            for k, v in value.items():
                value.pop(k)
                value[self.convert_name(k)] = v
        # setattr(element, self.convert_name(key), value)
        return

    def convert_name(self, name):
        if name:
            new_name = ''
            tmp = ''
            for ch in name:
                if ch.isupper():
                    tmp += ch
                    continue
                if tmp:
                    ch = '_' + tmp.lower() + ch
                    tmp = ''
                new_name += ch
            if tmp:
                new_name += '_' + tmp.lower()
            if new_name.startswith('_'):
                new_name = new_name[1:]

            return new_name

    # generics
    def get_list(self, action, params, markers):
        try:
            body = self.make_request(action, params)
            footmark.log.debug('body= %s' % body)
            return self.parse_response(markers, body, self)
        except ServerException as e:
            footmark.log.error('%s' % e)
            raise self.ResponseError(e)
        except Exception as e:
            footmark.log.error('%s' % e)
            raise e

    def get_status(self, action, params):
        try:
            body = self.make_request(action, params)
            footmark.log.debug('body= %s' % body)
            body = json.loads(body, encoding='UTF-8')
            if body:
                return True
            return False
        except ServerException as e:
            footmark.log.error('%s' % e)
            raise e
        except Exception as e:
            footmark.log.error('%s' % e)
            raise e

    def get_object(self, action, params, obj):
        try:
            body = self.make_request(action, params)
            footmark.log.debug(body)
            markers = ["", obj]
            obj = self.parse_response(markers, body, self)
            if obj:
                return obj
            return None
        except ServerException as e:
            footmark.log.error('%s' % e)
            raise e
        except Exception as e:
            footmark.log.error('%s' % e)
            raise e

    def make_request_new(self, params):
        if not params:
            raise Exception("Request parameters should not be empty.")

        conn = client.AcsClient(self.acs_access_key_id, self.acs_secret_access_key, self.region, user_agent=self.user_agent)
        if not conn:
            footmark.log.error('%s %s' % ('Null AcsClient ', conn))
            raise self.FootmarkClientError('Null AcsClient ', conn)

        timeout = 200
        delay = 3
        if not isinstance(params, dict):
            raise Exception("Invalid request parameters: {0} should be a dict.".format(params))

        if not params.get('Action', params.get('action')):
            raise Exception("'Action' is required for this request.")

        while timeout > 0:
            request = self.import_request(params.get('Action', params.get('action')))
            request.set_accept_format('json')
            try:
                for k, v in params.items():
                    if hasattr(request, k):
                        getattr(request, k)(v)
                    else:
                        request.add_query_param(k[4:], v)
                return conn.do_action_with_exception(request)
            except ServerException as e:
                if str(e.error_code) == "SDK.ServerUnreachable" \
                        or str(e.message).__contains__("SDK.ServerUnreachable") \
                        or str(e.message).__contains__("Unable to connect server: timed out"):
                    time.sleep(delay)
                    timeout -= delay
                    continue
                raise e
            except Exception as e:
                raise e

        return None

    def get_list_new(self, params, markers):
        try:
            body = self.make_request_new(params)
            footmark.log.debug('body= %s' % body)
            return self.parse_response(markers, body, self)
        except ServerException as e:
            footmark.log.error('%s' % e)
            raise self.ResponseError(e)
        except Exception as e:
            footmark.log.error('%s' % e)
            raise e

    def get_status_new(self, params):
        try:
            body = self.make_request_new(params)
            footmark.log.debug('body= %s' % body)
            body = json.loads(body, encoding='UTF-8')
            if body:
                return True
            return False
        except ServerException as e:
            footmark.log.error('%s' % e)
            raise e
        except Exception as e:
            footmark.log.error('%s' % e)
            raise e

    def get_object_new(self, params, obj):
        try:
            body = self.make_request_new(params)
            footmark.log.debug(body)
            markers = ["", obj]
            obj = self.parse_response(markers, body, self)
            if obj:
                return obj
            return None
        except ServerException as e:
            footmark.log.error('%s' % e)
            raise e
        except Exception as e:
            footmark.log.error('%s' % e)
            raise e



# encoding: utf-8
"""
Represents a connection to the OSS service.
"""

from footmark.oss.connection import OSSConnection
import oss2
import yaml

GMT_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'


class Bucket(OSSConnection): 
    """ 
    Object Storage Services
    """

    def __init__(self, acs_access_key_id, acs_secret_access_key,
                 region, bucket_name, user_agent=None):
        """
        Init method to create a new bucket object.
        """
        super(Bucket, self).__init__(acs_access_key_id, acs_secret_access_key, region, bucket_name)
        self.bucket_name = bucket_name
        self.bucket = oss2.Bucket(self.auth, self.endpoint, self.bucket_name)
        # self.grant = self.bucket.get_bucket_acl()

    def __getattr__(self, name):
        if name in ('id', 'name'):
            return self.bucket_name
        if name in ('permission', 'acl', 'grant'):
            return self.bucket.get_bucket_acl().acl
        if name == 'location':
            return self.bucket.get_bucket_location().location
        raise AttributeError

    def __setattr__(self, name, value):
        if name in ('id', 'name'):
            self.bucket_name = value
        if name in ('permission', 'acl'):
            self.grant = value
        super(OSSConnection, self).__setattr__(name, value)

    def create(self, permission):
        """
        Create a New Bucket
        :type permission: str
        :param permission: This option lets the user set the canned permissions on the bucket that are created
        :return: Details of newly created bucket
        """
        if self.is_exist():
            raise Exception("Error, Bucket with same name already exist.")

        self.bucket.create_bucket(permission=permission).headers

        return self

    def is_exist(self):
        """
        Verify bucket is exist
        :return: boolean value representing bucket existence
        """
        return self.is_bucket_exist(self.bucket_name)

    def delete(self):
        """
        Delete a Bucket
        :return: Returns status of operation
        """
        return self.bucket.delete_bucket()

    def put_acl(self, permission):
        """
        Update bucket acl
        :type permission: str
        :param permission: New bucket acl
        :rtype: class
        :return: Return the current bucket
        """
        self.bucket.put_bucket_acl(permission)
        setattr(self, 'grant', permission)
        return self

    def put_object(self, key, data, overwrite=False, headers=None, progress_callback=None):
        """
        Upload data to an object
        :type key: str
        :param key: Object name in bucket
        :type headers: dict
        :param headers: Custom headers for GET operation 
        :type data: str
        :param data: The content will be uploaded
        :type overwrite: bool
        :param overwrite: Whether overwrite existing object content. 
        If it is false and the specified object type is Appendable Object, the operation will append new content to it.
        :type progress_callback: str
        :param progress_callback: The callback function of upload progress
        :rtype class
        :return: `PutObjectResult <oss2.models.PutObjectResult>`
        """
        if headers and not isinstance(headers, dict):
            headers = yaml.load(headers)

        if not overwrite:
            position = 0
            if self.is_object_exist(key):
                obj = self.get_object_info(key)
                if obj.type != 'Appendable':
                    raise Exception('The specified object {0} is not Appendable. If you want to append it,'
                                    ' please remove it or put a new Appendable Object.'.format(key))
                position = obj.size
            return self.bucket.append_object(key, position, data, headers=headers, progress_callback=progress_callback)

        return self.bucket.put_object(key=key, data=data, headers=headers, progress_callback=progress_callback).headers

    def put_object_from_file(self, key, filename, overwrite=False, headers=None, progress_callback=None):
        """
        Upload a local file to an object
        :type key: str
        :param key: Object name in bucket
        :type headers: dict
        :param headers: Custom headers for GET operation 
        :type filename: str
        :param filename: The name of file used to upload
        :type overwrite: bool
        :param overwrite: Whether overwrite existing object content. 
        If it is false and the specified object type is Appendable Object, the operation will append new content to it.
        :type progress_callback: str
        :param progress_callback: The callback function of upload progress
        :rtype class
        :return: `PutObjectResult <oss2.models.PutObjectResult>`
        """
        if headers and not isinstance(headers, dict):
            headers = yaml.load(headers)

        if not overwrite:
            position = 0
            if self.is_object_exist(key):
                obj = self.get_object_info(key)
                if obj.type != 'Appendable':
                    raise Exception('The specified object {0} is not Appendable. If you want to append it,'
                                    ' please remove it and put a new Appendable Object.'.format(key))
                position = obj.size
            return self.bucket.append_object(key, position, open(filename).read(), headers=headers, progress_callback=progress_callback)

        return self.bucket.put_object_from_file(key, filename, headers=headers, progress_callback=progress_callback)

    def put_object_acl(self, key, permission):
        """
        Update object acl
        :type permission: str
        :param permission: New object acl
        :rtype: class
        :return: `RequestResult <oss2.models.RequestResult>`
        """
        return self.bucket.put_object_acl(key, permission)

    def is_object_exist(self, key):
        """
        Verify if object exists in Bucket
        :type key: str
        :param key: object key to verify for existence
        :rtype: bool
        :return: boolean value representing object existence
        """
        return self.bucket.object_exists(key)

    def update_object_headers(self, key, headers):
        """
        Upload a object's headers info
        :type key: str
        :param key: Object name in bucket
        :type headers: dict
        :param headers: Custom headers for GET operation 
        :rtype class
        :return: `RequestResult <oss2.models.RequestResults>`
        """
        if headers and not isinstance(headers, dict):
            headers = yaml.load(headers)
        return self.bucket.update_object_meta(key, headers)

    def get_object(self, key, byte_range=None, headers=None, progress_callback=None):
        """
        Download a object's content
        :type key: str
        :param key: Object name to download in bucket
        :type headers: dict
        :param headers: Custom headers for GET operation 
        :type byte_range: str
        :param byte_range: The range of content that would be download.
        Its format like 1-100 that indicates range from one to hundred bytes of object.
        :type progress_callback: str
        :param progress_callback: The callback function of download progress
        :rtype str
        :return: Download Object content
        """
        if headers and not isinstance(headers, dict):
            headers = yaml.load(headers)
        return self.bucket.get_object(key, byte_range=byte_range, headers=headers,
                                      progress_callback=progress_callback)

    def get_object_to_file(self, key, filename, byte_range=None, headers=None, progress_callback=None):
        """
        Download a object to a specified file
        :type key: str
        :param key: Object name to download in bucket
        :type headers: dict
        :param headers: Custom headers for GET operation 
        :type filename: str
        :param filename: The name of file used to store object content
        :type byte_range: str
        :param byte_range: The range of content that would be download.
        Its format like 1-100 that indicates range from one to hundred bytes of object.
        :type progress_callback: str
        :param progress_callback: The callback function of download progress
        :rtype object
        :return: Download Object file
        """
        if headers and not isinstance(headers, dict):
            headers = yaml.load(headers)

        if byte_range:
            byte_range=str.split(byte_range, '-')

        return self.bucket.get_object_to_file(key, filename, byte_range=byte_range, headers=headers,
                                              progress_callback=progress_callback)

    def get_object_info(self, key):
        """
        Get an specified Bucket Object Info
        :type key: str
        :param key: object name to retrieve in bucket.
        :rtype class: <oss2.models.SimplifiedObjectInfo>
        :return one specified object info
        """
        max_keys = 500

        while True:
            objects = self.list_objects(prefix=key, max_keys=max_keys)
            for obj in objects:
                if key == obj.key:
                    return obj
            if len(objects) < max_keys:
                return None

    def list_objects(self, prefix='', marker="", max_keys=100):
        """
        Lists Bucket Objects Info
        :type str
        :param prefix: retrieving all objects that have one specified prefix
        :type marker: str
        :param marker: the key to start with when using list mode. Object keys are returned in alphabetical order, 
         starting with key after the marker in order.
        :type max_keys: int
        :param max_keys: Max number of results to return in list mode
        :rtype List class: <oss2.models.SimplifiedObjectInfo>
        :return list of retrieved objects
        """

        return self.bucket.list_objects(prefix=prefix, marker=marker, max_keys=max_keys).object_list

    def delete_object(self, key):
        """
        Delete Object in Bucket
        :type key: str
        :param key: Object name to delete in bucket
        :rtype: bool
        :return: Whether delete object successfully
        """

        return self.bucket.delete_object(key)

    def delete_objects(self, keys):
        """
        Delete Objects in Bucket
        :type keys: list
        :param keys: objects to delete in bucket
        :rtype: bool
        :return: whether delete objects successfully
        """

        return self.bucket.batch_delete_objects(keys)




"""
Represents an ECS Instance
"""
from footmark.ess.essobject import TaggedESSObject


class ScalingConfiguration(TaggedESSObject):
    """
    Represents an scaling configuration.
    """

    def __init__(self, connection=None):
        super(ScalingConfiguration, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Scaling Configuration:%s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.scaling_configuration_id
        if name == 'name':
            return self.scaling_configuration_name
        if name == 'group_id':
            return self.scaling_group_id
        if name in ('state', 'status'):
            return self.lifecycle_state
        raise AttributeError("Object {0} does not have attribute {1}".format(self.__repr__(), name))

    def __setattr__(self, name, value):
        if name == 'id':
            self.scaling_configuration_id = value
        if name == 'name':
            self.scaling_configuration_name = value
        if name == 'group_id':
            self.scaling_group_id = value
        if name == 'lifecycle_state':
            value = value.lower()
        if name in ('state', 'status'):
            self.lifecycle_state = value
        if name == 'tags' and value:
            v = {}
            for tag in value['tag']:
                if tag.get('tag_key'):
                    v[tag.get('tag_key')] = tag.get('tag_value', None)
            value = v
        super(TaggedESSObject, self).__setattr__(name, value)

    def _update(self, updated):
        self.__dict__.update(updated.__dict__)

    def update(self, validate=False):
        """
        Update the instance's state information by making a call to fetch
        the current instance attributes from the service.

        :type validate: bool
        :param validate: By default, if ECS returns no data about the
                         instance the update method returns quietly.  If
                         the validate param is True, however, it will
                         raise a ValueError exception if no data is
                         returned from ECS.
        """
        rs = self.connection.describe_configurations(self.scaling_group_id, [self.id])
        if len(rs) > 0:
            for r in rs:
                if r.id == self.id:
                    self._update(r)
        elif validate:
            raise ValueError('%s is not a valid Scaling Configuration ID' % self.id)
        return self.state

    def active(self):
        """
        Start the instance.
        """
        return self.connection.start_instances([self.id])

    def inactive(self, force=False):
        """
        Stop the instance

        :type force: bool
        :param force: Forces the instance to stop

        :rtype: list
        :return: A list of the instances stopped
        """
        return self.connection.stop_instances([self.id], force)

    def terminate(self):
        """
        Terminate the instance

        :type force: bool
        :param force: Forces the instance to terminate
        """
        return self.connection.terminate_configuration(self.id)



# encoding: utf-8
"""
Represents a connection to the SLB service.
"""

import time
import json
from footmark.connection import ACSQueryConnection
from footmark.slb.regioninfo import RegionInfo
from footmark.exception import SLBResponseError
from footmark.slb.slb import LoadBalancer, BackendServer, VServerGroup, LoadBalancerListener


class SLBConnection(ACSQueryConnection):
    SDKVersion = '2014-05-15'
    DefaultRegionId = 'cn-hangzhou'
    DefaultRegionName = u'杭州'.encode("UTF-8")
    ResponseError = SLBResponseError

    def __init__(self, acs_access_key_id=None, acs_secret_access_key=None,
                 region=None, sdk_version=None, security_token=None, user_agent=None):
        """
        Init method to create a new connection to SLB.
        """
        if not region:
            region = RegionInfo(self, self.DefaultRegionName,
                                self.DefaultRegionId)
        self.region = region
        if sdk_version:
            self.SDKVersion = sdk_version

        self.SLBSDK = 'aliyunsdkslb.request.v' + self.SDKVersion.replace('-', '')

        super(SLBConnection, self).__init__(acs_access_key_id,
                                            acs_secret_access_key,
                                            self.region, self.SLBSDK, security_token, user_agent=user_agent)
    
    def describe_vserver_group_attribute(self, vserver_group_id):
        """
        describe vserver group
        :type vserver_group_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :return: return the vserver group object
        """
        params = {}
        self.build_list_params(params, vserver_group_id, 'VServerGroupId')

        return self.get_object('DescribeVServerGroupAttribute', params, VServerGroup)
    
    def  describe_vserver_groups(self, load_balancer_id):
        """
        describe vserver group
        :type vserver_group_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :return: return the vserver group object
        """
        params = {}
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        
        return self.get_object('DescribeVServerGroups', params, ['VServerGroups' ,VServerGroup])
    
    def create_vserver_group(self, load_balancer_id, vserver_group_name, backend_servers):
        """
        create vserver group
        :type vserver_group_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :type load_balancer_id: string
        :param load_balancer_id: Unique identifier for load balancer
        :type backend_servers: list
        :param backend_servers:Backend server list
        :return: return the vserver group object
        """
        params = {}
        backend_serverlist = []
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, vserver_group_name, 'VServerGroupName')
        for servers in backend_servers:
            backend_serverlist.append({
                'ServerId': servers['instance_id'],
                'Port': servers['port'],
                'Weight': servers['weight']
            })
        self.build_list_params(params, backend_serverlist, 'BackendServers')
        
        return self.get_object('CreateVServerGroup', params, VServerGroup)
    
    def set_vserver_group_attribute(self, vserver_group_id, vserver_group_name = '', backend_servers = []):
        """
        set vserver group attribute
        :type vserver_group_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :type load_balancer_id: string
        :param load_balancer_id: Unique identifier for load balancer
        :type backend_servers: list
        :param backend_servers:Backend server list
        :return: return the vserver group object
        """
        params = {}
        backend_serverlist = []
        self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        if vserver_group_name:
            self.build_list_params(params, vserver_group_name, 'VServerGroupName')
        if backend_servers:
            for servers in backend_servers:
                backend_serverlist.append({
                    'ServerId': servers['instance_id'],
                    'Port': servers['port'],
                    'Weight': servers['weight']
                })
            self.build_list_params(params, backend_serverlist, 'BackendServers')
        
        return self.get_object('SetVServerGroupAttribute', params, VServerGroup)
    
    def add_vserver_group_backend_servers(self, vserver_group_id, backend_servers):
        """
        add vserver group backend servers
        :type vserver_group_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :type backend_servers: list
        :param backend_servers:Backend server list
        :return: return the vserver group object
        """
        params = {}
        backend_serverlist = []
        self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        self.build_list_params(params, backend_servers, 'BackendServers')
        for servers in backend_servers:
            backend_serverlist.append({
                'ServerId': servers['instance_id'],
                'Port': servers['port'],
                'Weight': servers['weight']
            })
        self.build_list_params(params, backend_serverlist, 'BackendServers')
        
        return self.get_object('AddVServerGroupBackendServers', params, VServerGroup)
    
    def remove_vserver_group_backend_servers(self, vserver_group_id, backend_servers):
        """
        remove vserver group backend servers
        :type vserver_group_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :type backend_servers: list
        :param backend_servers:Backend server list
        :return: return the vserver group object
        """
        params = {}
        backend_serverlist = []
        self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        for servers in backend_servers:
            backend_serverlist.append({
                'ServerId': servers['instance_id'],
                'Port': servers['port']
            })
        self.build_list_params(params, backend_serverlist, 'BackendServers')
        
        return self.get_object('RemoveVServerGroupBackendServers', params, VServerGroup)
    
    def modify_vserver_group_backend_servers(self, vserver_group_id, old_backend_servers = [], new_backend_servers = []):
        """
        modify vserver group backend servers
        :type vserver_group_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :type old_backend_servers: list
        :param old_backend_servers:Old backend server list
        :type new_backend_servers: list
        :param new_backend_servers:new backend server list
        :return: return the vserver group object
        """
        params = {}
        old_backend_serverlist = []
        new_backend_serverlist = []
        self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        if old_backend_servers:
            for servers in old_backend_servers:
                old_backend_serverlist.append({
                    'ServerId': servers['instance_id'],
                    'Port': servers['port']
            })
            self.build_list_params(params, old_backend_serverlist, 'OldBackendServers')
        if new_backend_servers:
            for servers in new_backend_servers:
                new_backend_serverlist.append({
                    'ServerId': servers['instance_id'],
                    'Port': servers['port'],
                    'Weight': servers['weight']
            })
            self.build_list_params(params, new_backend_serverlist, 'NewBackendServers')
        
        return self.get_object('ModifyVServerGroupBackendServers', params, VServerGroup)
    
    def delete_vserver_group(self, vserver_group_id):
        """
        delete vserver group
        :type load_balancer_id: string
        :param vserver_group_id: Unique identifier for the virtual server group
        :return: return bool
        """
        params = {}
        self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        
        return self.get_status('DeleteVServerGroup', params)    

    def create_load_balancer(self, load_balancer_name=None, address_type=None, vswitch_id=None, client_token=None,
                             internet_charge_type=None, master_zone_id=None, slave_zone_id=None, bandwidth=None):
        """
        Creates a Server Load Balancer
        :type load_balancer_name: string
        :param load_balancer_name: Name to the server load balancer
        :type address_type: string
        :param address_type:  Address type. value: internet or intranet
        :type vswitch_id: string
        :param vswitch_id: The vswitch id of the VPC instance. This option is invalid if address_type parameter is
         provided as internet.
        :type internet_charge_type: string
        :param internet_charge_type: Charging mode for the public network instance
         Value: paybybandwidth or paybytraffic
        :type master_zone_id: string
        :param master_zone_id: Name of of availability zones to enable on this SLB
        :type slave_zone_id: string
        :param slave_zone_id: Name of of availability zones to enable on this SLB
        :type bandwidth: string
        :param bandwidth: Bandwidth peak of the public network instance charged per fixed bandwidth
        :return: return the created load balancer details
        """

        params = {}

        if load_balancer_name:
            self.build_list_params(params, load_balancer_name, 'LoadBalancerName')
        if address_type:
            self.build_list_params(params, address_type, 'AddressType')
        if vswitch_id:
            self.build_list_params(params, vswitch_id, 'VSwitchId')
        if internet_charge_type:
            self.build_list_params(params, internet_charge_type, 'InternetChargeType')
        if master_zone_id:
            self.build_list_params(params, master_zone_id, 'MasterZoneId')
        if slave_zone_id:
            self.build_list_params(params, slave_zone_id, 'SlaveZoneId')
        if bandwidth:
            self.build_list_params(params, bandwidth, 'Bandwidth')
        if client_token:
            self.build_list_params(params, client_token, 'ClientToken')
        slb = self.get_object('CreateLoadBalancer', params, LoadBalancer)
        return self.describe_load_balancer_attribute(slb.id)

    def add_listeners(self, load_balancer_id, purge_listener=None, listeners=None):
        """
        Add Listeners to existing ServerLoadBalancer
        :type load_balancer_id: str
        :param load_balancer_id: Id of ServerLoadBalancer
        :type purge_listener: bool
        :param purge_listener:  Whether to remove existing Listener or not
        :type listeners: dict
        :param listeners: List of ports/protocols for this SLB to listen on
        :return: returns RequestId id of request
        """
        params = {}
        results = []
        deleted_listener = []
        changed = False

        try:
            # find out all listeners of the load balancer
            self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
            slb_details = self.get_status('DescribeLoadBalancerAttribute', params)

            # if purge_listener is true then delete existing listeners
            if purge_listener:
                if slb_details:
                    if len(slb_details[u'ListenerPortsAndProtocal'][u'ListenerPortAndProtocal']) > 0:
                        for slb_listener in slb_details[u'ListenerPortsAndProtocal'][u'ListenerPortAndProtocal']:
                            params = {}
                            self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
                            self.build_list_params(params, slb_listener[u'ListenerPort'], 'ListenerPort')
                            response = self.get_status('DeleteLoadBalancerListener', params)
                            deleted_listener.append(response)
                            changed = True
                            
            # add listeners to load balancer
            if listeners:
                for listener in listeners:
                    if listener:
                        if 'protocol' in listener:
                            protocol = str(listener['protocol']).lower()
                            # Add HTTP Listener to Load Balancer
                            if protocol in ['http']:
                                listener_result = self.create_load_balancer_http_listener(load_balancer_id, listener)
                                if listener_result:                                    
                                    results.append({"http_listener_result": listener_result[1]})
                                    # modify changed param according to listener result
                                    if changed is False:
                                        changed = listener_result[0]

                            # Add HTTPS Listener to Load Balancer
                            elif protocol in ['https']:
                                listener_result = self.create_load_balancer_https_listener(load_balancer_id, listener)
                                if listener_result:
                                    results.append({"https_listener_result": listener_result[1]})
                                    # modify changed param according to listener result
                                    if changed is False:
                                        changed = listener_result[0]

                            # Add TCP Listener to Load Balancer
                            elif protocol in ['tcp']:
                                listener_result = self.create_load_balancer_tcp_listener(load_balancer_id, listener)
                                if listener_result:
                                    results.append({"tcp_listener_result": listener_result[1]})
                                    # modify changed param according to listener result
                                    if changed is False:
                                        changed = listener_result[0]

                            # Add UDP Listener to Load Balancer
                            elif protocol in ['udp']:
                                listener_result = self.create_load_balancer_udp_listener(load_balancer_id, listener)
                                if listener_result:
                                    results.append({"udp_listener_result": listener_result[1]})
                                    # modify changed param according to listener result
                                    if changed is False:
                                        changed = listener_result[0]
                            else:
                                results.append({"Error Message": "Invalid Listener Protocol " + listener['protocol']})

        except Exception as ex:
            error_code = ex.error_code
            error_msg = ex.message
            results.append({"Error Code": error_code, "Error Message": error_msg})

        return changed, results

    def create_load_balancer_listener(self, load_balancer_id,\
                                            protocol,\
                                            listener_port,\
                                            backend_server_port,\
                                            bandwidth=None,\
                                            sticky_session='',\
                                            health_check='',\
                                            scheduler='',\
                                            sticky_session_type='',\
                                            cookie_timeout=None,\
                                            cookie='',\
                                            health_check_domain='',\
                                            health_check_uri='',\
                                            health_check_connect_port=None,\
                                            healthy_threshold=None,\
                                            unhealthy_threshold=None,\
                                            health_check_timeout=None,\
                                            health_check_interval=None,\
                                            health_check_http_code='',\
                                            vserver_group_id='',\
                                            gzip='',\
                                            server_certificate_id='',\
                                            master_slave_server_group_id='',\
                                            persistence_timeout=None,\
                                            health_check_connect_timeout=None,\
                                            xforwarded_for = 'on',\
                                            ca_certificate_id=''):
        """
        Create HTTP Listener; create Listeners based on the HTTP protocol for the Server Load Balancer instance,
        including policies and health check configurations based on the HTTP protocol
        :type load_balancer_id: string
        :param load_balancer_id: load balance id
        :type listener_port: int
        :param listener_port: Server Load Balancer instance frontend port. Value: 1-65535
        :type backend_server_port: int
        :param backend_server_port: Server Load Balancer instance backend port. Value: 1-65535
        :type bandwidth: int
        :param bandwidth: Listener bandwidth. Value: -1 / 1-1000 Mbps
        :type sticky_session: string
        :param sticky_session: Whether to open the Gzip compression
        :type protocol: string
        :param protocol: type of listener to create
        :type health_check: string
        :param health_check: Whether to enable health check
        :type health_check: string
        :param health_check: Whether to enable health check
        :type: scheduler: string
        :param: scheduler: Scheduling algorithm. Value: wrr / wlc / rr Default value: wrr
        :type: sticky_session_type: string
        :param: sticky_session_type: Mode for handling the cookie.
        :type: cookie_timeout: int
        :param: cookie_timeout: Cookie timeout.
        :type: cookie: string
        :param: cookie: The cookie configured on the server
        :type: health_check_domain: string
        :param: health_check_domain: Domain name used for health check.
        :type: health_check_uri: string
        :param: health_check_uri: URI used for health check
        :type: health_check_connect_port: int
        :param: health_check_connect_port: Port used for health check.
        :type: healthy_threshold: int
        :param: healthy_threshold: Threshold determining the result of the health check is success.
        :type: unhealthy_threshold: int
        :param: unhealthy_threshold: Threshold determining the result of the health check is fail. 
        :type: health_check_timeout: int
        :param: health_check_timeout: Maximum timeout of each health check response. 
        :type: health_check_http_code: string
        :param: health_check_http_code: Regular health check HTTP status code. Multiple codes are segmented by ",".
        :type: health_check_interval: int
        :param: health_check_interval: Time interval of health checks.
        :type: vserver_group_id: string
        :param: vserver_group_id: Virtual server group ID.
        :type: gzip: string
        :param: gzip: whether open Gzip compression
        :type: server_certificate_id: string
        :param: server_certificate_id: Server certificate ID.
        :type: master_slave_server_group_id: string
        :param: master_slave_server_group_id: Master standby server group ID
        :type: persistence_timeout: string
        :param: persistence_timeout: Timeout time for connection persistence.
        :type: health_check_connect_timeout: int
        :param: health_check_connect_timeout: Health check connection timeout.
        :type: xforwarded_for: int
        :param: xforwarded_for: Whether open access to the actual IP of visitors through X-Forwarded-For
        :type: ca_certificate_id: string
        :param: ca_certificate_id: CA certificate ID
        :return: returns request status
        """

        params = {}
        results = []
        changed = False
        listener_type_dic = dict(http="CreateLoadBalancerHTTPListener",\
                                 https= "CreateLoadBalancerHTTPSListener",\
                                 tcp="CreateLoadBalancerTCPListener",\
                                 udp="CreateLoadBalancerUDPListener")
        key = listener_type_dic.get(protocol, '')

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')
        self.build_list_params(params, backend_server_port, 'BackendServerPort')
        if bandwidth:
            self.build_list_params(params, bandwidth, 'Bandwidth')
        if sticky_session:
            self.build_list_params(params, sticky_session, 'StickySession')
        if health_check:
            self.build_list_params(params, health_check, 'HealthCheck')
        if scheduler:
            self.build_list_params(params, scheduler, 'Scheduler')
        if gzip:
            self.build_list_params(params, gzip, 'Gzip')
        if server_certificate_id:
            self.build_list_params(params, server_certificate_id, 'ServerCertificateId')
        if xforwarded_for:
            self.build_list_params(params, 'on', 'XForwardedFor')
        if sticky_session_type:
            self.build_list_params(params, sticky_session_type, 'StickySessionType')
        if cookie_timeout:
            self.build_list_params(params, cookie_timeout, 'CookieTimeout')
        if cookie:
            self.build_list_params(params, cookie, 'Cookie')
        if health_check_domain:
            self.build_list_params(params, health_check_domain, 'HealthCheckDomain')
        if health_check_uri:
            self.build_list_params(params, health_check_uri, 'HealthCheckURI')
        if health_check_connect_port:
            self.build_list_params(params, health_check_connect_port, 'HealthCheckConnectPort')
        if healthy_threshold:
            self.build_list_params(params, healthy_threshold, 'HealthyThreshold')
        if unhealthy_threshold:
            self.build_list_params(params, unhealthy_threshold, 'UnhealthyThreshold')
        if health_check_timeout:
            self.build_list_params(params, health_check_timeout, 'HealthCheckTimeout')
        if health_check_interval:
            self.build_list_params(params, health_check_interval, 'HealthCheckInterval')
        if health_check_http_code:
            self.build_list_params(params, health_check_http_code, 'HealthCheckHttpCode')
        if vserver_group_id:
            self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        if master_slave_server_group_id:
            self.build_list_params(params, master_slave_server_group_id, 'MasterSlaveServerGroupId')
        if persistence_timeout:
            self.build_list_params(params, persistence_timeout, 'PersistenceTimeout')
        if health_check_connect_timeout:
            self.build_list_params(params, health_check_connect_timeout, 'HealthCheckConnectTimeout')
        if ca_certificate_id:
            self.build_list_params(params, ca_certificate_id, 'CACertificateId')

        if self.get_status(key, params):
            return self.start_load_balancer_listener(load_balancer_id, listener_port)
        return False

    def set_listener_access_control_status(self, load_balancer_id, listener_port, access_control_status):
        """
        set listener access control status
        :type load_balancer_id: string
        :param load_balancer_id: Load balancer instance  id
        :type listener_port: int
        :param listener_port: Load balancer instance  frontend port. Value: 1-65535
        :type access_control_status: string
        :param load_balancer_id: Whether or not access control is enabled. open_white_list indicates the white list access control function is enabled.
        :return: returns bool
        """
        params = {}

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')
        self.build_list_params(params, access_control_status, 'AccessControlStatus')
        return self.get_status('SetListenerAccessControlStatus', params)

    def set_listener_attribute(self, load_balancer_id,\
                                    listener_port,\
                                    bandwidth,\
                                    protocol,\
                                    sticky_session='',\
                                    health_check='',\
                                    scheduler='',\
                                    sticky_session_type='',\
                                    cookie_timeout=None,\
                                    cookie='',\
                                    health_check_domain='',\
                                    health_check_uri='',\
                                    health_check_connect_port=None,\
                                    healthy_threshold=None,\
                                    unhealthy_threshold=None,\
                                    health_check_timeout=None,\
                                    health_check_interval=None,\
                                    health_check_http_code='',\
                                    vserver_group_id='',\
                                    gzip='',\
                                    server_certificate_id='',\
                                    master_slave_server_group_id='',\
                                    persistence_timeout=None,\
                                    health_check_connect_timeout=None,\
                                    ca_certificate_id='',\
                                    syn_proxy='',\
                                    health_check_type='',\
                                    vserver_group='',\
                                    master_slave_server_group='',
                                    xforwarded_for='on'):
        """
        set listener attribute
        :type load_balancer_id: string
        :param load_balancer_id: load balance id
        :type listener_port: int
        :param listener_port: Server Load Balancer instance frontend port. Value: 1-65535
        :type bandwidth: int
        :param bandwidth: Listener bandwidth. Value: -1 / 1-1000 Mbps
        :type sticky_session: string
        :param sticky_session: Whether to open the Gzip compression
        :type protocol: string
        :param protocol: type of listener to create
        :type health_check: string
        :param health_check: Whether to enable health check
        :type: scheduler: string
        :param: scheduler: Scheduling algorithm. Value: wrr / wlc / rr Default value: wrr
        :type: sticky_session_type: string
        :param: sticky_session_type: Mode for handling the cookie.
        :type: cookie_timeout: int
        :param: cookie_timeout: Cookie timeout.
        :type: cookie: string
        :param: cookie: The cookie configured on the server
        :type: health_check_domain: string
        :param: health_check_domain: Domain name used for health check.
        :type: health_check_uri: string
        :param: health_check_uri: URI used for health check
        :type: health_check_connect_port: int
        :param: health_check_connect_port: Port used for health check.
        :type: healthy_threshold: int
        :param: healthy_threshold: Threshold determining the result of the health check is success.
        :type: unhealthy_threshold: int
        :param: unhealthy_threshold: Threshold determining the result of the health check is fail.
        :type: health_check_timeout: int
        :param: health_check_timeout: Maximum timeout of each health check response.
        :type: health_check_http_code: string
        :param: health_check_http_code: Regular health check HTTP status code. Multiple codes are segmented by ",".
        :type: health_check_interval: int
        :param: health_check_interval: Time interval of health checks.
        :type: vserver_group_id: string
        :param: vserver_group_id: Virtual server group ID.
        :type: gzip: string
        :param: gzip: whether open Gzip compression
        :type: server_certificate_id: string
        :param: server_certificate_id: Server certificate ID.
        :type: master_slave_server_group_id: string
        :param: master_slave_server_group_id: Master standby server group ID
        :type: persistence_timeout: string
        :param: persistence_timeout: Timeout time for connection persistence.
        :type: health_check_connect_timeout: int
        :param: health_check_connect_timeout: Health check connection timeout.
        :type: xforwarded_for: int
        :param: xforwarded_for: Whether open access to the actual IP of visitors through X-Forwarded-For
        :type: ca_certificate_id: string
        :param: ca_certificate_id: CA certificate ID
        :type: syn_proxy: string
        :param: syn_proxy: CA certificate ID
        :type: health_check_type: string
        :param: health_check_type: health check type
        :type: vserver_group: string
        :param: vserver_group: Whether to use a virtual server group
        :type: master_slave_server_group: string
        :param: master_slave_server_group: Whether to use the primary and secondary server groups
        :return: returns request status
        """
        params = {}
        listener_type_dic = dict(http="SetLoadBalancerHTTPListenerAttribute",\
                                 https= "SetLoadBalancerHTTPSListenerAttribute",\
                                 tcp="SetLoadBalancerTCPListenerAttribute",\
                                 udp="SetLoadBalancerUDPListenerAttribute")
        key = listener_type_dic.get(protocol, '')

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')
        self.build_list_params(params, bandwidth, 'Bandwidth')
        if listener_port:
            self.build_list_params(params, listener_port, 'ListenerPort')
        if health_check:
            self.build_list_params(params, health_check, 'HealthCheck')
        if sticky_session:
            self.build_list_params(params, sticky_session, 'StickySession')
        if server_certificate_id:
            self.build_list_params(params, server_certificate_id, 'ServerCertificateId')
        if scheduler:
            self.build_list_params(params, scheduler, 'Scheduler')
        if xforwarded_for:
            self.build_list_params(params, 'on', 'XForwardedFor')
        if sticky_session_type:
            self.build_list_params(params, sticky_session_type, 'StickySessionType')
        if cookie_timeout:
            self.build_list_params(params, cookie_timeout, 'CookieTimeout')
        if cookie:
            self.build_list_params(params, cookie, 'Cookie')
        if health_check_domain:
            self.build_list_params(params, health_check_domain, 'HealthCheckDomain')
        if health_check_uri:
            self.build_list_params(params, health_check_uri, 'HealthCheckURI')
        if health_check_connect_port:
            self.build_list_params(params, health_check_connect_port, 'HealthCheckConnectPort')
        if healthy_threshold:
            self.build_list_params(params, healthy_threshold, 'HealthyThreshold')
        if unhealthy_threshold:
            self.build_list_params(params, unhealthy_threshold, 'UnhealthyThreshold')
        if health_check_timeout:
            self.build_list_params(params, health_check_timeout, 'HealthCheckTimeout')
        if health_check_interval:
            self.build_list_params(params, health_check_interval, 'HealthCheckInterval')
        if health_check_http_code:
            self.build_list_params(params, health_check_http_code, 'HealthCheckHttpCode')
        if vserver_group:
            self.build_list_params(params, vserver_group, 'VServerGroup')
        if vserver_group_id:
            self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        if master_slave_server_group_id:
            self.build_list_params(params, master_slave_server_group_id, 'MasterSlaveServerGroupId')
        if master_slave_server_group:
            self.build_list_params(params, master_slave_server_group, 'MasterSlaveServerGroup')
        if ca_certificate_id:
            self.build_list_params(params, ca_certificate_id, 'CACertificateId')
        if syn_proxy:
            self.build_list_params(params, syn_proxy, 'SynProxy')
        if health_check_type:
            self.build_list_params(params, health_check_type, 'HealthCheckType')
        if gzip:
            self.build_list_params(params, gzip, 'Gzip')
        if persistence_timeout:
            self.build_list_params(params, persistence_timeout, 'PersistenceTimeout')
        if health_check_connect_timeout:
            self.build_list_params(params, health_check_connect_timeout, 'HealthCheckConnectTimeout')

        return self.get_status(key, params)

    def delete_load_balancer_listener(self, load_balancer_id, listener_port):
        """
        delete load balance listener
        :type load_balancer_id: string
        :param load_balancer_id: Load balancer instance  id
        :type listener_port: string
        :param listener_port: Load balancer instance  frontend port. Value: 1-65535
        :return: returns bool
        """
        params = {}
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')

        return self.get_status('DeleteLoadBalancerListener', params)

    def start_load_balancer_listener(self, load_balancer_id, listener_port):
        """
        start load balance listener
        :type load_balancer_id: string
        :param load_balancer_id: Load balancer instance  id
        :type listener_port: string
        :param listener_port: Load balancer instance  frontend port. Value: 1-65535
        :return: returns bool
        """
        params = {}
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')

        return self.get_status('StartLoadBalancerListener', params)

    def stop_load_balancer_listener(self, load_balancer_id, listener_port):
        """
        stop load balance listener
        :type load_balancer_id: string
        :param load_balancer_id: Load balancer instance  id
        :type listener_port: string
        :param listener_port: Load balancer instance  frontend port. Value: 1-65535
        :return: returns bool
        """
        params = {}
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')

        return self.get_status('StopLoadBalancerListener', params)

    def add_listener_white_list_item(self, load_balancer_id, listener_port, source_items):
        """
        add load balance listener white list item
        :type load_balancer_id: string
        :param load_balancer_id: Load balancer instance  id
        :type listener_port: string
        :param listener_port: Load balancer instance  frontend port. Value: 1-65535
        :type source_items: string
        :param source_items: Access control list
        :return: returns bool
        """
        params = {}
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')
        self.build_list_params(params, source_items, 'SourceItems')

        return self.get_status('AddListenerWhiteListItem', params)

    def remove_listener_white_list_item(self, load_balancer_id, listener_port, source_items):
        """
        remove load balance listener white list item
        :type load_balancer_id: string
        :param load_balancer_id: Load balancer instance  id
        :type listener_port: string
        :param listener_port: Load balancer instance  frontend port. Value: 1-65535
        :type source_items: string
        :param source_items: Access control list
        :return: returns bool
        """
        params = {}
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')
        self.build_list_params(params, source_items, 'SourceItems')

        return self.get_status('RemoveListenerWhiteListItem', params)

    def describe_load_balancer_listener_attribute(self, load_balancer_id, listener_port, protocol):
        """
        describe load balancer listener attribute
        :type load_balancer_id: string
        :param load_balancer_id: Load balancer instance  id
        :type listener_port: string
        :param listener_port: Load balancer instance  frontend port. Value: 1-65535
        :return: obj: object of listener
        """
        action = "DescribeLoadBalancer"
        if protocol is None or len(str(protocol))<=0:
            lb = self.describe_load_balancer_attribute(load_balancer_id)
            if lb is None:
                return None
            for pp in lb.listener_ports_and_protocol.get("listener_port_and_protocol"):
                if pp.get("listener_port") == listener_port:
                    protocol = pp.get("listener_protocol")
                    break

        action += str.upper(protocol) + "ListenerAttribute"

        params = {}

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, listener_port, 'ListenerPort')
        try:
            obj = self.get_object(action, params, LoadBalancerListener)
        except Exception as e:
            obj = None
        return obj

    def add_backend_servers(self, load_balancer_id, backend_servers=None):
        """
        Add BackendServer to existing LoadBalancer
        :type load_balancer_id: str
        :param load_balancer_id: ID of server load balancer
        :type backend_servers: list
        :param backend_servers: list of dictionary containing server Id and weight of  backend server instance
        :return: return changed status, current_backend_servers and message with descriptive information
        """

        params = {}
        results = []
        current_backend_servers = []
        changed = False

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')

        backend_servers_list = []

        for backend_server in backend_servers:
            backend_servers_list.append({"ServerId": backend_server['server_id'],
                                         "Weight": str(backend_server['weight'])})

        backend_servers_json = json.dumps(backend_servers_list)
        self.build_list_params(params, backend_servers_json, 'BackendServers')

        return self.get_list('AddBackendServers', params, ["BackendServers", BackendServer])
    
    def purge_add_backend_server(self, load_balancer_id, instance_ids=None, purge_instance_ids=None):
        """
        Remove existing Instances or Backend Server and Add new instances or Backend Server to Load Balancer
        :type load_balancer_id: str
        :param load_balancer_id: Id of ServerLoadBalancer
        :type instance_ids:list
        :param instance_ids: Id of Instances or Backend Server
        :type purge_instance_ids: bool
        :param purge_instance_ids: Whether to remove existing Instances or Backend Servers
        :return: Returns Id of newly added Load Balancer
        """
        params = {}
        results = []
        instances = []
        changed = False

        try:
            # List all Backend Servers
            self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
            response = self.get_status('DescribeLoadBalancerAttribute', params)
            for instance in response[u'BackendServers'][u'BackendServer']:
                # append id of all Backend Servers to list
                instances.append(str(instance[u'ServerId']))

            # Remove instances only when purge_instance_ids is True
            if len(instances) > 0 and (purge_instance_ids is True):
                # Remove all Backend Servers
                response = self.remove_backend_servers(load_balancer_id=load_balancer_id, backend_server_ids=instances)
                if 'error' in (''.join(str(response))).lower():
                    results.append(response[2])
                else:
                    results.append(response[2][0])
                    changed = True

            # Add Backend Server to Load Balancer
            if instance_ids:
                if len(instance_ids) > 0:
                    backend_servers = []
                    for backend_server_id in instance_ids:
                        backend_servers.append({"server_id": backend_server_id, "weight": 100})

                    response = self.add_backend_servers(load_balancer_id, backend_servers)
                    if 'error' in (''.join(str(response))).lower():
                        results.append({"backend_server_result": response[2]})
                    else:
                        results.append({"backend_server_result": response[1][0]})
                        changed = True
        except Exception as ex:
            error_code = ex.error_code
            error_msg = ex.message
            results.append({"Error Code": error_code, "Error Message": error_msg})

        return changed, results    

    def remove_backend_servers(self, load_balancer_id=None, backend_server_ids=None):
        """
        :type load_balancer_id: str
        :param load_balancer_id: ID of server load balancer
        :type backend_server_ids: list
        :param backend_server_ids: list of IDs of backend server instance
        :return: return changed status, current_backend_servers and message with descriptive information
        """
        params = {}

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')

        backend_servers_json = json.dumps(backend_server_ids)

        self.build_list_params(params, backend_servers_json, 'BackendServers')

        return self.get_list('RemoveBackendServers', params, ["BackendServers", BackendServer])

    def set_backend_servers(self, load_balancer_id=None, backend_servers=None):
        """
        Set Backend Server to Load Balancer
        :type load_balancer_id: str
        :param load_balancer_id: ID of server load balancer
        :type backend_servers: list
        :param backend_servers: list of dictionary containing server Id and weight of  backend server instance
        :return: return changed status, current_backend_servers and message with descriptive information
        """

        params = {}
        results = []
        current_backend_servers = []
        changed = False

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')

        backend_servers_list = []

        for backend_server in backend_servers:
            backend_servers_list.append({"ServerId": backend_server['server_id'],
                                         "Weight": str(backend_server['weight'])})

        backend_servers_json = json.dumps(backend_servers_list)
        self.build_list_params(params, backend_servers_json, 'BackendServers')

        return self.get_list('SetBackendServers', params, ["BackendServers", BackendServer])

    def describe_backend_servers_health_status(self, load_balancer_id=None, port=None):
        """
        :type load_balancer_id: str
        :param load_balancer_id: ID of server load balancer
        :type port: list
        :param port: list of Ports used by the Server Load Balancer instance frontend for health check
        :return: return backend servers with health status and message with descriptive information
        """
        params = {}
        results = []
        backend_servers_health_status = []

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')

        if port:
            self.build_list_params(params, port, 'ListenerPort')

        return self.get_list('DescribeHealthStatus', params, ["BackendServers", BackendServer])

    def set_load_balancer_status(self, load_balancer_id, load_balancer_status):
        """
        Method added to Set Load Balancer Status
        :type load_balancer_id: List
        :param load_balancer_id: ID of server load balancer
        :type load_balancer_status: String
        :param load_balancer_status: Status of an Server Load Balancer instance
            Value: inactive | active
        :return: return name of the operating interface, which is
            specified in the system
        """
        params = {}
       
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')        
        self.build_list_params(params, load_balancer_status, 'LoadBalancerStatus')
        
        return self.get_status('SetLoadBalancerStatus', params)

    def set_load_balancer_name(self, load_balancer_id, load_balancer_name):
        """
        Set name or alias to the ServerLoadBalancer
        Method added to Set Load Balancer Name
        :type load_balancer_id: str
        :param load_balancer_id: ID of a Server Load Balancer instance
        :type load_balancer_id: str
        :param load_balancer_name: Displayed name of an Server Load Balancer instance. When the parameter is not
         specified, an instance name is allocated by the system by default.
        :return: returns the request_id of request
        """
        changed = False
        results = []
        params = {}
        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        self.build_list_params(params, load_balancer_name, 'LoadBalancerName')
        return self.get_status('SetLoadBalancerName', params)

    def delete_load_balancer(self, slb_id):
        """
        Method added to Delete Load Balancer
        :type slb_id: string
        :param slb_id: Id of the server load balancer
        :return: Return status of Operation
        """
        results = []
        params = {}

        self.build_list_params(params, slb_id, 'LoadBalancerId')
        return self.get_status('DeleteLoadBalancer', params)   
        
    def modify_slb_internet_spec(self, load_balancer_id, internet_charge_type=None, bandwidth=None):
        """
        Modify internet specifications of existing LoadBalancer, like internet_charge_type or bandwidth
        :type load_balancer_id: str
        :param load_balancer_id: The unique ID of an Server Load Balancer instance
        :type internet_charge_type: str
        :param internet_charge_type: Charging mode for the public network instance
        :type bandwidth: int
        :param bandwidth: Bandwidth peak of the public network instance charged per fixed bandwidth
        :return: returns the request_id of request
        """
        
        results = []
     
        params = {}  

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        if internet_charge_type:
            self.build_list_params(params, internet_charge_type, 'InternetChargeType')
        if bandwidth:
            self.build_list_params(params, bandwidth, 'Bandwidth')
        return self.get_status('ModifyLoadBalancerInternetSpec', params)
    

    def describe_load_balancer_attribute(self, load_balancer_id):
        """
        Describe attributes of Load Balancer
        :type load_balancer_id: string
        :param load_balancer_id: id of the load balancer
        :return: load balance attributes in dictionary format if found else None
        """

        params = {}

        self.build_list_params(params, load_balancer_id, 'LoadBalancerId')

        return self.get_object('DescribeLoadBalancerAttribute', params, LoadBalancer)
    
    def describe_load_balancers(self, load_balancer_id = None, load_balancer_name = None):
        """
        Describe Load Balancers
        :type load_balancer_id: string
        :param load_balancer_id: id of the load balancer
        :type load_balancer_name: string
        :param load_balancer_name: name of the load balancer
        :return: load balance in dictionary format if found else None
        """

        params = {}
        if load_balancer_id:
            self.build_list_params(params, load_balancer_id, 'LoadBalancerId')
        if load_balancer_name:
            self.build_list_params(params, load_balancer_name, 'LoadBalancerName')
        return self.get_list('DescribeLoadBalancers', params,  ['LoadBalancers', LoadBalancer])

    def add_vservergroup_backend_server(self, vserver_group_id, backend_servers):
        """
        Add a back-end server in a virtual server group, add a set of back-end servers to a specific virtual server
            group in the SLB,
        and return a list of back-end servers in that virtual server group.
        :type vserver_group_id: string
        :param vserver_group_id: The unique identifier for the virtual server group
        :param backend_servers:  - List of hash/dictionary of backend servers to add in
          - '[{"key":"value", "key":"value"}]', keys allowed:
            - server_id (required:true, description: Unique id of Instance to add)
            - port (required:true, description: The back-end server using the port, range: 1-65535)
            - weight (required:true; default: 100, description: Weight of the backend server, in the range of 1-100 )
        :return: VServerGroupId	String	The unique identifier for the virtual server group
                 BackendServers	List	Array format, returns the operation is successful,
                 the virtual server group all the back-end server list, the list of elements in the structure see
                 BackendServer
        """
        params = {}
        changed = False
        results = []
        backend_serverlist = []       
        if vserver_group_id:
            self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        if backend_servers:
            for servers in backend_servers:
                backend_serverlist.append({'ServerId': servers['server_id'],
                                           'Port': servers['port'],
                                           'Weight': servers['weight']})
                
            self.build_list_params(params, json.dumps(backend_serverlist), 'BackendServers')
        try:
            results = self.get_status('AddVServerGroupBackendServers', params)           
            changed = True
        except Exception as ex:
            error_code = str(ex.error_code)
            error_msg = str(ex.message)
            results.append("Error Code:" + error_code + " ,Error Message:" + error_msg)

        return changed, results

    def remove_vserver_group_backend_server(self, vserver_group_id, purge_backend_servers):
        """        
        Method to Remove Vserver Group Backend server
        :type vserver_group_id: string
        :param vserver_group_id: Virtual server group Id
        :type purge_backend_servers:  List of hash/dictionary
        :param purge_backend_servers:
          - List of hash/dictionary of backend servers to remove
          - '[{"key":"value", "key":"value"}]', keys allowed:
            - server_id (required:true, description: Unique id of Instance to remove)
            - port (required:true, description: The back-end server using the port, range: 1-65535)          
                               
        :return: it return public parameters with ,VServerGroupId The unique identifier for the virtual server.
                 and BackendServers Array format, list of back-end servers in the virtual server group.
                 The structure of the elements in the list is detailed in BackendServer
        """
        
        params = {}
        results = []
        changed = False
        backend_serverlist = []
           
        if vserver_group_id:
            self.build_list_params(params, vserver_group_id, 'VServerGroupId')
        if purge_backend_servers:
            for servers in purge_backend_servers:
                backend_serverlist.append({'ServerId': servers['server_id'], 'Port': servers['port']})
            self.build_list_params(params, json.dumps(backend_serverlist), 'BackendServers')

        try: 
            results = self.get_status('RemoveVServerGroupBackendServers', params)
            changed = True 
        except Exception as ex:
            error_code = str(ex.error_code)
            error_msg = str(ex.message)
            results.append({"Error Code": error_code, "Error Message": error_msg})

        return changed, results

    def modify_vserver_group_backend_server(self, vserver_group_id, purge_backend_servers, backend_servers):
        '''
        Modify VServer Group Backend Server
        :type vserver_group_id: string
        :param vserver_group_id:Virtual server group Id
        :type purge_backend_servers:  List of hash/dictionary
        :param purge_backend_servers:
          - List of hash/dictionary of backend servers to add in
          - '[{"key":"value", "key":"value"}]', keys allowed:
            - server_id (required:true, description: Unique id of Instance to add)
            - port (required:true, description: The back-end server using the port, range: 1-65535)
        :type backend_servers:  List of hash/dictionary
        :param backend_servers:
          - List of hash/dictionary of backend servers to add in
          - '[{"key":"value", "key":"value"}]', keys allowed:
            - server_id (required:true, description: Unique id of Instance to add)
            - port (required:true, description: The back-end server using the port, range: 1-65535)
            - weight (required:true; default: 100, description: Weight of the backend server, in the range of 1-100 )

        :return: Change the virtual back-end servers in the server group, in a particular SLB virtual server
                 group by adding / deleting the back-end server to replace the current server group, the group returned
                 to the virtual server back-end server list.
        '''
        params = {}
        results = []
        set_results = []
        add_results = []
        set_server_attribute = []
        add_backend_server = []
        delete_backend_servers = []
        delete_backend_servers_id = []
        filter_backend_servers_id = []
        final_backend_servers_id = []
        changed = False
        serverid_param = 'server_id'
        try: 

            self.build_list_params(params, vserver_group_id, 'VServerGroupId')
            result_all_backend_servers = self.get_status('DescribeVServerGroupAttribute', params)
            all_backend_servers = result_all_backend_servers['BackendServers']['BackendServer']
            if all_backend_servers:
                for purge_backend_server in purge_backend_servers:                    
                    for all_backend_server in all_backend_servers:
                        if purge_backend_server[serverid_param] in all_backend_server['ServerId']:
                            delete_backend_servers.append(purge_backend_server)
                            delete_backend_servers_id.append(purge_backend_server[serverid_param])
                            break
                
                for backend_server in backend_servers:
                    flag = False
                    for all_backend_server in all_backend_servers:
                        if backend_server[serverid_param] == all_backend_server['ServerId'] \
                                and backend_server[serverid_param] not in delete_backend_servers_id:
                            set_server_attribute.append(backend_server)
                            flag = True
                            break                       
                    if not flag:   
                        add_backend_server.append(backend_server)
            else:
                 add_backend_server.append(backend_servers[0])             
           
            if delete_backend_servers:
               changed, result = self.remove_vserver_group_backend_server(vserver_group_id, delete_backend_servers)
            
            if set_server_attribute:
                changed, set_results = self.set_vservergroup_attribute(vserver_group_id, vserver_group_name=None,
                                                                       backend_servers=set_server_attribute)
    
            if add_backend_server:
                changed, add_results = self.add_vservergroup_backend_server(vserver_group_id, add_backend_server)
            changed = True
            
            if set_results:
                results = set_results
                filter_backend_servers_id = set_results['BackendServers']['BackendServer']
            
            if add_results:
                results = add_results
                filter_backend_servers_id += add_results['BackendServers']['BackendServer']

            for backend_server in backend_servers:
                flag = False
                for filter_backend_servers in filter_backend_servers_id:
                    if filter_backend_servers["ServerId"] == backend_server[serverid_param] and\
                                    filter_backend_servers["Port"] == backend_server["port"] and not flag:
                        flag = True
                        final_backend_servers_id.append(filter_backend_servers)
            if final_backend_servers_id:
                results['BackendServers']['BackendServer'] = final_backend_servers_id
            if 'VServerGroupName' in results:
                del results['VServerGroupName']
        except Exception as ex:
            error_code = str(ex.error_code)
            error_msg = str(ex.message)
            results.append({"Error Code": error_code, "Error Message": error_msg})

        return changed, results

    def describe_vservergroup_backendserver(self, vserver_group_id, backend_servers):
        """
        describe vserver group backend server      
        :type vserver_group_id: string
        :param vserver_group_id: The unique identifier for the virtual server group
        :type backend_servers:  List of hash/dictionary
        :param backend_servers:
          - List of hash/dictionary of backend servers to add in
          - '[{"key":"value", "key":"value"}]', keys allowed:
            - server_id (required:true, description: Unique id of Instance to add)
            - port (required:true, description: The back-end server using the port, range: 1-65535)
            - weight (required:true; default: 100, description: Weight of the backend server, in the range of 1-100 )
        :return: VServerGroupId	String	The unique identifier for the virtual server group
                 BackendServers	List	Array format, returns the operation is successful,
                 the virtual server group all the back-end server list, the list of elements in the structure
                 see BackendServer
        """
        changed_flag = True
        results = []   
        try:
            changed_vsg, result_vsgs = self.describe_vservergroup_attributes(vserver_group_id=vserver_group_id)
            if result_vsgs and changed_vsg:
                for backend_server in backend_servers:
                    final_check =False
                    for result_vsg in result_vsgs["BackendServers"]["BackendServer"]:
                        if backend_server["server_id"] == result_vsg["ServerId"] and backend_server["port"] \
                                == result_vsg["Port"]:
                            final_check = True  
                            break              
                    if not final_check: 
                        results.append(str(backend_server["server_id"])+" ECS with "
                                                                        "the port "+str(backend_server["port"])+" not "
                                                                        "match to perform operation")
                        changed_flag = False
            else:
                changed_flag = False
                results = result_vsgs
        except Exception as ex:
            error_code = str(ex.error_code)
            error_msg = str(ex.message)
            results.append("Error Code:" + error_code + " ,Error Message:" + error_msg)

        return changed_flag, results

    def describe_vservergroup_backendserver_to_add(self, vserver_group_id, backend_servers):
        """
        describe vserver group backend server to add
        and return a list of back-end servers in that virtual server group.
        :type vserver_group_id: string
        :param vserver_group_id: The unique identifier for the virtual server group
        :type backend_servers:  List of hash/dictionary
        :param backend_servers:
          - List of hash/dictionary of backend servers to add in
          - '[{"key":"value", "key":"value"}]', keys allowed:
            - server_id (required:true, description: Unique id of Instance to add)
            - port (required:true, description: The back-end server using the port, range: 1-65535)
            - weight (required:true; default: 100, description: Weight of the backend server, in the range of 1-100 )
        :return: VServerGroupId	String	The unique identifier for the virtual server group
                 BackendServers	List	Array format, returns the operation is successful,
                 the virtual server group all the back-end server list, the list of elements in the structure
                 see BackendServer
        """
        changed_flag = True
        results = []   
        try:
            changed_vsg, result_vsgs = self.describe_vservergroup_attributes(vserver_group_id=vserver_group_id)
            if result_vsgs and changed_vsg:
                for backend_server in backend_servers:
                    changed_flag = True
                    for result_vsg in result_vsgs["BackendServers"]["BackendServer"]:
                        if (str(backend_server["server_id"]) == str(result_vsg["ServerId"])) \
                                and (str(backend_server["port"]) == str(result_vsg["Port"])):
                            changed_flag = False
                            break

                    if changed_flag is False:
                        results.append(str(backend_server["server_id"])+" "
                                                                        "ECS with port "+str(backend_server["port"])+" "
                                                                        "is already present")
            else:
                changed_flag =False
                results = result_vsgs
        except Exception as ex:
            error_code = str(ex.error_code)
            error_msg = str(ex.message)
            results.append("Error Code:" + error_code + " ,Error Message:" + error_msg)

        return changed_flag, results

    # endregion






"""
Represents an ESS Scaling Rule
"""
from footmark.ess.essobject import TaggedESSObject


class ScalingRule(TaggedESSObject):
    """
    Represents an scaling configuration.
    """

    def __init__(self, connection=None):
        super(ScalingRule, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Scaling Rule: %s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.scaling_rule_id
        if name in ('name', 'rule_name'):
            return self.scaling_rule_name
        if name == 'ari':
            return self.scaling_rule_ari
        if name == 'group_id':
            return self.scaling_group_id
        raise AttributeError("Object {0} does not have attribute {1}".format(self.__repr__(), name))

    def __setattr__(self, name, value):
        if name in ('id', 'rule_id'):
            self.scaling_rule_id = value
        if name in ('name', 'rule_name'):
            self.scaling_rule_name = value
        if name == 'ari':
            self.scaling_rule_ari = value
        if name == 'group_id':
            self.scaling_group_id = value
        super(TaggedESSObject, self).__setattr__(name, value)

    def _update(self, updated):
        self.__dict__.update(updated.__dict__)

    def update(self, validate=False):
        """
        Update the instance's state information by making a call to fetch
        the current instance attributes from the service.scaling_rule_ids=None, scaling_rule_names=None

        :type validate: bool
        :param validate: By default, if ECS returns no data about the
                         instance the update method returns quietly.  If
                         the validate param is True, however, it will
                         raise a ValueError exception if no data is
                         returned from ECS.
        """
        rs = self.connection.describe_rules(scaling_rule_ids=[self.id])
        if len(rs) > 0:
            for r in rs:
                if r.id == self.id:
                    self._update(r)
        elif validate:
            raise ValueError('%s is not a valid Scaling Rule ID' % self.id)
        return self

    def modify(self, adjustment_type=None, adjustment_value=None, name=None, cooldown=None):
        """
        modify the scaling rule

        :type force: bool
        :param force: Forces the instance to stop

        :rtype: bool
        :return: result of modifying
        """
        update = False
        if adjustment_type and self.adjustment_type != adjustment_type:
            update = True
        if adjustment_value and self.adjustment_value != adjustment_value:
            update = True
        if name and self.name != name:
            update = True
        if cooldown is not None and self.cooldown != cooldown:
            update = True

        if not update:
            return False

        return self.connection.modify_rule(self.id, adjustment_type=adjustment_type, adjustment_value=adjustment_value,
                                           name=name, cooldown=cooldown)

    def terminate(self):
        """
        Terminate the rule
        """
        return self.connection.terminate_rule(self.id)



from footmark.regioninfo import RegionInfo


class RDSRegionInfo(RegionInfo):
    """
    Represents an RDS Region
    """

    def __init__(self, connection=None, name=None, id=None,
                 connection_cls=None):
        from footmark.rds.connection import RDSConnection
        super(RDSRegionInfo, self).__init__(connection, name, id,
                                            RDSConnection)



"""
This module provides an interface to the Elastic Compute Service (ECS) service from Alicloud.
"""
from footmark.vpc.connection import VPCConnection
from footmark.regioninfo import get_regions


def regions(**kw_params):
    """
    Get all available regions for the ECS service.
    You may pass any of the arguments accepted by the ECSConnection
    object's constructor as keyword arguments and they will be
    passed along to the ECSConnection object.

    :rtype: list
    :return: A list of :class:`footmark.ecs.regioninfo.RegionInfo`
    """
    return get_regions('ecs', connection_cls=VPCConnection)


def connect_to_region(region_id, **kw_params):
    """
    Given a valid region name, return a
    :class:`footmark.ecs.connection.ECSConnection`.
    Any additional parameters after the region_name are passed on to
    the connect method of the region object.

    :type: str
    :param region_id: The ID of the region to connect to.

    :rtype: :class:`footmark.ecs.connection.ECSConnection` or ``None``
    :return: A connection to the given region, or None if an invalid region
             name is given
    """
    return VPCConnection(region=region_id, **kw_params)


def get_region(region_id, **kw_params):
    """
    Find and return a :class:`footmark.ecs.regioninfo.RegionInfo` object
    given a region name.

    :type: str
    :param: The name of the region.

    :rtype: :class:`footmark.ecs.regioninfo.RegionInfo`
    :return: The RegionInfo object for the given region or None if
             an invalid region name is provided.
    """
    for region in regions(**kw_params):
        if region.id == region_id:
            return region
    return None



"""
Represents an ESS Scheduled Task
"""
from footmark.ess.essobject import TaggedESSObject


class ScheduledTask(TaggedESSObject):
    """
    Represents an scheduled task.
    """

    def __init__(self, connection=None):
        super(ScheduledTask, self).__init__(connection)
        self.tags = {}

    def __repr__(self):
        return 'Scheduled Task: %s' % self.id

    def __getattr__(self, name):
        if name == 'id':
            return self.scheduled_task_id
        if name == 'name':
            return self.scheduled_task_name
        if name in ('scaling_rule_ari', 'rule_ari', 'ari'):
            return self.scheduled_action
        if name == 'launch_expiration':
            return self.launch_expiration_time
        if name == 'enabled':
            return self.task_enabled
        raise AttributeError("Object {0} does not have attribute {1}".format(self.__repr__(), name))

    def __setattr__(self, name, value):
        if name == 'id':
            self.scheduled_task_id = value
        if name == 'name':
            self.scheduled_task_name = value
        if name in ('scaling_rule_ari', 'rule_ari', 'ari'):
            self.scheduled_action = value
        if name == 'launch_expiration':
            self.launch_expiration_time = value
        if name == 'enabled':
            self.task_enabled = value
        super(TaggedESSObject, self).__setattr__(name, value)

    def _update(self, updated):
        self.__dict__.update(updated.__dict__)

    def update(self, validate=False):
        """
        Update the instance's state information by making a call to fetch
        the current instance attributes from the service.scaling_group_ids=None, scaling_group_names=None

        :type validate: bool
        :param validate: By default, if ECS returns no data about the
                         instance the update method returns quietly.  If
                         the validate param is True, however, it will
                         raise a ValueError exception if no data is
                         returned from ECS.
        """
        rs = self.connection.describe_scheduled_tasks(scheduled_task_ids=[self.id])
        if len(rs) > 0:
            for r in rs:
                if r.id == self.id:
                    self._update(r)
        elif validate:
            raise ValueError('%s is not a valid Scheduled Task ID' % self.id)
        return self

    def modify(self, scaling_rule_ari=None, launch_time=None, name=None, description=None,
               launch_expiration_time=None, recurrence_type=None, recurrence_value=None,
               recurrence_end_time=None, task_enabled=None):
        """
        Modify the scheduled task
        """
        update = False
        if scaling_rule_ari and self.scaling_rule_ari != scaling_rule_ari:
            update = True
        if launch_time and self.launch_time != launch_time:
            update = True
        if name and self.name != name:
            update = True
        if description and getattr(self, 'description', None) and self.description != description:
            update = True
        if launch_expiration_time is not None and self.launch_expiration_time != launch_expiration_time:
            update = True
        if recurrence_type and self.recurrence_type != recurrence_type:
            update = True
        if recurrence_value and self.recurrence_value != recurrence_value:
            update = True
        if recurrence_end_time and self.recurrence_end_time != recurrence_end_time:
            update = True
        if task_enabled is not None and self.task_enabled != task_enabled:
            update = True

        if not update:
            return False

        return self.connection.modify_scheduled_task(self.id, scaling_rule_ari=scaling_rule_ari, launch_time=launch_time,
                                                     name=name, description=description, launch_expiration_time=launch_expiration_time,
                                                     recurrence_type=recurrence_type, recurrence_value=recurrence_value,
                                                     recurrence_end_time=recurrence_end_time, task_enabled=task_enabled)

    def terminate(self):
        """
        Terminate the scheduled task

        """
        return self.connection.terminate_scheduled_task(self.id)



#!/usr/bin/env python
# Always prefer setuptools over distutils
from codecs import open
from os import path

try:
    from setuptools import setup

    extra = dict(test_suite="tests.test.suite", include_package_data=True)
except ImportError:
    from distutils.core import setup

    extra = {}

PACKAGE = "footmark"
NAME = "footmark"
DESCRIPTION = "A Python interface to Aliyun Web Services"
AUTHOR = "xiaozhu"
AUTHOR_EMAIL = "heguimin36@163.com"
URL = ""
VERSION = __import__(PACKAGE).__version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=NAME,

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=VERSION,

    description=DESCRIPTION,
    long_description=long_description,

    # The project's main homepage.
    url=URL,

    # Author details
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=["footmark", "footmark.ecs", "footmark.slb", "footmark.vpc", "footmark.rds", "footmark.ess",
              "footmark.oss", "footmark.pyami"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['aliyun-python-sdk-ecs>=2.1.0', 'aliyun-python-sdk-slb>=2.0.21', 'aliyun-python-sdk-core>=2.3.2',
                      'aliyun-python-sdk-rds>=2.1.0', 'aliyun-python-sdk-ess>=2.1.3', 'oss2>=2.3.3', 'importlib']
)



# #!/usr/bin/env python
# # import sys
# # sys.path.append("../../..")
#
# from tests.unit import OSSMockServiceTestCase
# from footmark.oss.bucket import Bucket
# import json
#
# CREATE_FOLDER = '''
# '''
#
# class Key():
#     key = "myKey"
#
# class LIST_OBJECTS():
#     object_list=[Key()]
#
#
# DELETE_OBJECTS = '''
# '''
#
# CREATE_BUCKET = '''
# '''
#
# DELETE_BUCKET = '''
# '''
#
# SIMPLE_UPLOAD = '''
# '''
#
#
# class TestCreateBucket(OSSMockServiceTestCase):
#     connection_class = Bucket
#
#     permission = "private"
#
#     def default_body(self):
#         return CREATE_BUCKET
#
#     def test_create_bucket(self):
#         self.set_http_response(status_code=200)
#         changed, result = self.service_connection.create_bucket(permission=self.permission)
#         if changed is True:
#             self.assertEqual(result[0], 'Bucket Created Successfully')
#
#
# class TestDeleteBucket(OSSMockServiceTestCase):
#     connection_class = Bucket
#
#     def default_body(self):
#         return DELETE_BUCKET
#
#     def test_delete_bucket(self):
#         self.set_http_response(status_code=200)
#         changed, result = self.service_connection.delete_bucket()
#         if changed is True:
#             self.assertEqual(result[0], 'Bucket Deleted Successfully')
#
#
# class TestSimpleUpload(OSSMockServiceTestCase):
#     connection_class = Bucket
#
#     bucket_name = "tushar"
#     src = "D:\local_file.txt"
#     file_name = "remote.txt"
#     headers = {
#         "Content-Type": "text/html",
#         "Content-Encoding": "MD5"
#
#     }
#     metadata = {
#         "x-oss-meta-key": "value"
#     }
#     expiration = "30"
#     encrypt = None
#     overwrite = None
#
#     def default_body(self):
#         return SIMPLE_UPLOAD
#
#     def test_simple_upload(self):
#         self.set_http_response(status_code=200)
#         changed, result = self.service_connection.simple_upload(
#             expiration=self.expiration, headers=self.headers, encrypt=self.encrypt, metadata=self.metadata,
#             overwrite=self.overwrite, src=self.src, file_name=self.file_name)
#         if changed is True:
#             self.assertEqual(result[0], 'File uploaded Successfully')
#
#
# class TestCreateBucketFolder(OSSMockServiceTestCase):
#     connection_class = Bucket
#
#     folder_name = 'test/'
#
#     def default_body(self):
#         return CREATE_FOLDER
#
#     def test_create_folder(self):
#         self.set_http_response(status_code=200)
#         changed, result = self.service_connection.create_folder(self.folder_name)
#         self.assertEqual(changed, True)
#
#
# class TestListBucketObjects(OSSMockServiceTestCase):
#     connection_class = Bucket
#
#     marker = 'my'
#     max_keys = 100
#
#     def default_body(self):
#         return LIST_OBJECTS
#
#     def test_list_objects(self):
#         self.set_http_response(status_code=200)
#         keys, result = self.service_connection.list_bucket_objects(marker=self.marker, max_keys=self.max_keys)
#         self.assertEqual(result[0], 'Bucket objects retrieved successfully')
#
# class TestDeleteBucketObjects(OSSMockServiceTestCase):
#     connection_class = Bucket
#
#     objects_to_delete= ["myKey", "myDocs/"]
#
#     def default_body(self):
#         return DELETE_OBJECTS
#
#     def test_delete_objects(self):
#         self.set_http_response(status_code=200)
#         changed, results = self.service_connection.delete_bucket_objects(objects=self.objects_to_delete)
#         self.assertEqual(changed, True)
#
#
#
#
#
